from __future__ import annotations
import os
from threading import Thread
import datetime
import json
import numpy as np
import pandas as pd
from scipy.spatial.distance import euclidean
from fastdtw import fastdtw
from rich.console import Console
from rich.panel import Panel
from rich.text import Text
import matplotlib
import matplotlib.pyplot as plt
import plotly.graph_objects as go
import plotly.offline
import plotly.express as px
from plotly.subplots import make_subplots

from ftio.freq.freq_data import FreqData
from ftio.freq.helper import format_plot
from ftio.plot.units import set_unit
from ftio.freq.freq_html import create_html

# import plotly.io as pio

matplotlib.rcParams["backend"] = "TkAgg"

class FreqPlot:
    """For plotting the result of ftio"""

    def __init__(self, argv):
        self.render = "dynamic"
        self.plot_engine = "plotly"
        self.dtw = True
        self.dominant = []
        self.recon = False
        self.psd = False
        if not isinstance(argv, bool):
            D1 = []
            D2 = []
            D3 = []
            j = 0
            self.check = 0
            self.modes = ["write_async", "write_sync", "read_async", "read_sync"]
            self.s = []
            mode = "write_async"
            if isinstance(argv, list):
                if len(argv) <= 1:
                    self.paths = ["."]
                else:
                    self.paths = []
                    for i in range(1, len(argv)):
                        if "sync_" in str(argv[i]):
                            mode = str(argv[i])
                        elif "static" in str(argv[i]) or "dynamic" in str(argv[i]):
                            self.render = str(argv[i])
                        else:
                            self.paths.append(str(argv[i]))
            else:
                self.paths = [argv]

            print("\033[1;32m--------------------------------------------\n\033[1;32m")
            self.Print_info(argv[0])
            print("\n\033[1;35mPlot mode is %s\033[1;0m\n" % (mode))
            for path in self.paths:
                print(
                    "\n\033[1;34mLoading folder(%i,%i): %s\033[1;0m"
                    % (self.paths.index(path) + 1, len(self.paths), path)
                )
                # print ("\033[1;34m   "+'-'*len("Current folder(%i,%i): %s" % (self.paths.index(path) + 1, len(self.paths),path))+"\033[1;0m")
                if "DFT.json" not in path[-8:]:  # not single file
                    # path = path[:path.rfind("/")]
                    if path[-1] == "/":
                        path = path[:-1]
                    for root, _, files in os.walk(path):
                        for file in sorted(files, key=len):
                            if "DFT.json" in file[-8:]:
                                file = os.path.join(root, file)
                                print(" '-> Current file: " + file)
                                f = open(file)
                                data = json.load(f)
                                mode = self.Check_Mode(data, mode)
                                D1.append(pd.DataFrame(data=data.get(mode).get("data")))
                                D2.append(
                                    pd.DataFrame(
                                        data=data.get(mode).get("settings"), index=[j]
                                    )
                                )
                                D3.append(
                                    pd.DataFrame(data=data.get(mode).get("original"))
                                )
                                D1[j] = pd.concat(
                                    [
                                        D1[j],
                                        pd.DataFrame(
                                            {
                                                "ranks": [D2[j]["ranks"][j]]
                                                * D2[j]["N"][j],
                                                "k": [
                                                    x for x in range(0, D2[j]["N"][j])
                                                ],
                                                "freq": [
                                                    x
                                                    / (D2[j]["N"][j] * D2[j]["T_s"][j])
                                                    for x in range(0, D2[j]["N"][j])
                                                ],
                                                "T": [
                                                    (
                                                        (
                                                            D2[j]["N"][j]
                                                            * D2[j]["T_s"][j]
                                                        )
                                                        / x
                                                        if x > 0
                                                        else 0
                                                    )
                                                    for x in range(0, D2[j]["N"][j])
                                                ],
                                            }
                                        ),
                                    ],
                                    axis=1,
                                )
                                if "b" in D3[j]:
                                    D3[j] = pd.concat(
                                        [
                                            D3[j],
                                            pd.DataFrame(
                                                {
                                                    "ranks": [D2[j]["ranks"][j]]
                                                    * len(D3[j]["b"])
                                                }
                                            ),
                                        ],
                                        axis=1,
                                    )
                                j += 1
                        break  # no recusive walk
                else:
                    file = path
                    print("| -> Current file: " + file)
                    f = open(file)
                    data = json.load(f)
                    mode = self.Check_Mode(data, mode)
                    D1.append(pd.DataFrame(data=data.get(mode).get("data")))
                    D2.append(
                        pd.DataFrame(data=data.get(mode).get("settings"), index=[j])
                    )
                    D3.append(pd.DataFrame(data=data.get(mode).get("original")))
                    D1[j] = pd.concat(
                        [
                            D1[j],
                            pd.DataFrame(
                                {
                                    "ranks": [D2[j]["ranks"][j]] * D2[j]["N"][j],
                                    "k": [x for x in range(0, D2[j]["N"][j])],
                                    "freq": [
                                        x / (D2[j]["N"][j] * D2[j]["T_s"][j])
                                        for x in range(0, D2[j]["N"][j])
                                    ],
                                    "T": [
                                        (
                                            (D2[j]["N"][j] * D2[j]["T_s"][j]) / x
                                            if x > 0
                                            else 0
                                        )
                                        for x in range(0, D2[j]["N"][j])
                                    ],
                                }
                            ),
                        ],
                        axis=1,
                    )
                    if "b" in D3[j]:
                        D3[j] = pd.concat(
                            [
                                D3[j],
                                pd.DataFrame(
                                    {"ranks": [D2[j]["ranks"][j]] * len(D3[j]["b"])}
                                ),
                            ],
                            axis=1,
                        )
                    j += 1

            # print("--------------------------------------------\n")
            self.n = len(self.s)
            self.D = FreqData(
                pd.concat(D1, ignore_index=True),
                pd.concat(D2, ignore_index=True),
                pd.concat(D3, ignore_index=True),
            )

    def add_df(self, n, D1, D2, D3, D4=[]):
        self.n = n
        self.D = FreqData(
            pd.concat(D1, ignore_index=True),
            pd.concat(D2, ignore_index=True),
            pd.concat(D3, ignore_index=True),
        )
        if isinstance(D4, list) and D4:  # and not D4.empty:
            self.dominant = pd.concat(D4, ignore_index=True)

    def set(self, dict):
        for prop, value in dict.items():
            if prop in "render":
                self.render = value
            elif prop in "plot_engine":
                self.plot_engine = value
            elif prop in "dtw":
                self.dtw = value
            elif prop in "reconstruction":
                self.recon = value
            elif prop in "psd":
                self.psd = value

    def Check_Mode(self, data, mode):
        if self.check == 0:
            if not self.modes:
                print("no valid mode")
                exit(0)
            if not data.get(mode)["data"]:
                self.modes.remove(mode)
                mode = self.Check_Mode(data, self.modes[0])
            else:
                self.check = 1
                if len(self.modes) < 4:
                    print("\033[1;35mMode adjusted %s\033[1;0m" % (mode))
        return mode

    def Print_info(self, s):
        name = text[text.rfind("/") + 1 : -3].capitalize()
        console = Console()
        title = Panel(
            Text(name, justify="center"),
            style="bold white on cyan",
            border_style="white",
            title_align="left",
        )
        text = "\n[cyan]Author:[/] Ahmad Tarraf\n"
        text += f"[cyan]Date:[/]   {str(datetime.date.today())}\n"
        text += f"[cyan]Version:[/]   {1.0}\n"
        text += f"[cyan]License:[/]   --\n"
        # console.print(Panel(Group(title,text), style="white", border_style='blue'))
        console.print(title)
        console.print(text)

    def Plot(self):
        if "mat" not in self.plot_engine and "plotly" not in self.plot_engine:
            return

        # Settings
        console = Console()
        all_or_10 = False
        # template = "plotly_dark"
        # conf = {"toImageButtonOptions": {"format": "svg", "scale": 1}}
        conf = {"toImageButtonOptions": {"format": "png", "scale": 4}}
        template = "plotly"
        width = 1100
        height = 500  # 600
        font_settings = {
            "family": "Courier New, monospace",
            "size": 24,
            "color": "black",
        }
        colors = px.colors.qualitative.Plotly
        colors.pop(1)

        # Init
        top = []
        f = []
        bar_plot = go.Figure()
        color_counter = 0

        if "dark" in template:
            color_bar = "white"
        else:
            color_bar = "black"
        # https://plotly.com/python/renders/
        if self.render == "dynamic":
            # pio.renders.default = "browser"
            slider_cond = False
            visible = "legendonly"
        else:
            # pio.renders.default ="png"
            slider_cond = False
            visible = True
        console.print(
            f"\n[underline cyan]Ploting:[/]\n"
            f"[cyan]Plot render:[/] {self.render}\n"
            f"[cyan]Plot engine:[/] {self.plot_engine}\n"
            f"[cyan]DTW calculation:[/] {self.dtw}\n"
        )

        def Scatter(**kwargs):
            if self.render == "dynamic":
                return go.Scatter(kwargs)
            else:
                return go.Scattergl(kwargs)

        ranks = np.sort(pd.unique(self.D.settings_df["ranks"]))
        for r in ranks:
            index_set = self.D.settings_df["ranks"].isin([r])
            index_data = self.D.data_df["ranks"].isin([r])
            if not self.D.original_df.empty:
                index_original = self.D.original_df["ranks"].isin([r])

            samples = np.arange(0, self.D.settings_df["N"][index_set].values)
            amp     = np.array(self.D.data_df[index_data]["A"])
            freq    = np.array(self.D.data_df[index_data]["freq"])
            phi     = np.array(self.D.data_df[index_data]["phi"])
            found   = False

            # reconstruct time
            time = (
                self.D.settings_df[index_set]["t_start"].values
                + samples * self.D.settings_df[index_set]["T_s"].values
            )

            if self.plot_engine == "plotly":
                f.append(go.Figure())
            else:
                bar_plot = plt.figure(figsize=(10, 4))
                f1 = plt.figure(figsize=(10, 4))

            limit = (
                self.D.data_df[index_data].index.min()
                + self.D.settings_df[index_set]["N"].values / 2
            )
            if "conf" in self.D.data_df:
                sorted_ref = self.D.data_df[index_data].sort_values("conf")
                dominant = self.dominant[self.dominant["ranks"].isin([r])]["k"].values
                # dominant=[0,7,14, 21, 28, 35,42,49, 56]
                if all_or_10:
                    dominant = np.concatenate(
                        (dominant, self.D.settings_df[index_set]["N"][0] - dominant)
                    )
            else:
                sorted_ref = self.D.data_df[index_data].sort_values("A")
                dominant = sorted_ref.tail(3).index

            top_3 = sorted_ref.tail(7).index
            top_3 = top_3[top_3 < limit[0]]
            top.append(top_3[-2:-1][0])
            if not all_or_10:
                # top_x = self.D.data_df[index_data].sort_values('A').tail(20).index
                top_x = self.D.data_df[index_data].sort_values("A").tail(5).index
                top_x = top_x[top_x < limit[0]]

            name_dominant = ""
            # set the unit and order
            unit, order = set_unit(self.D.data_df[index_data]["b_sampled"])
            #
            # unit, order = set_unit(self.D.data_df[index_data]["b_sampled"]*1000) #compatibility with old version (For BWLIMIT)
            for k in samples:
                x = (
                    (1 / len(samples))
                    * amp[k]
                    * np.cos(
                        2
                        * np.pi
                        * samples
                        * k
                        / (self.D.settings_df[index_set]["N"].values)
                        + phi[k]
                    )
                )
                if k == 0:
                    sum_all_components = x
                    sum_dominant = np.zeros(x.size)
                    # recon with DC offset
                    if self.recon:
                        sum_top_2 = x
                        sum_top_3 = x
                        sum_top_5 = x
                        sum_top_10 = x
                    dominant_X1 = x
                    dominant_X2 = x
                    dominant_k1 = top_3[-2]
                    dominant_k2 = top_3[-3]

                else:
                    sum_all_components = sum_all_components + x
                    if k in dominant:
                        sum_dominant = sum_dominant + x
                        found = True
                    if self.recon:
                        if k in sorted_ref.tail(3).index:
                            sum_top_2 = sum_top_2 + x
                        if k in sorted_ref.tail(5).index:
                            sum_top_3 = sum_top_3 + x
                        if k in sorted_ref.tail(9).index:
                            sum_top_5 = sum_top_5 + x
                        if k in sorted_ref.tail(19).index:
                            sum_top_10 = sum_top_10 + x

                    if k == dominant_k1:
                        dominant_X1 = (
                            dominant_X1 + 2 * x if not all_or_10 else dominant_X1 + x
                        )
                    if k == dominant_k2:
                        dominant_X2 = (
                            dominant_X2 + 2 * x if not all_or_10 else dominant_X2 + x
                        )

                length = len(samples)
                if "mat" in self.plot_engine:
                    if all_or_10:
                        s = "%.1e*cos(2\u03C0*%.1f*t%+.2f)" % (
                            order * amp[k] / length,
                            k,
                            phi[k],
                        )
                        plt.plot(time, x * order, linewidth=0.7, label="_nolegend_")
                    else:
                        if self.D.data_df[index_data].index[k] in top_x:
                            if k == 0 or k == len(samples) / 2:
                                a = 1
                            else:
                                a = 2

                            if round(freq[k], 2) > 0:
                                s = f"{a / length * order*amp[k]:.1e}*cos(2\u03C0*{freq[k]:.2f}*t+{phi[k]:.2f})"
                            else:
                                s = f"{a / length * order*amp[k]:.1e}*cos(2\u03C0*{freq[k]:.2e}*t+{phi[k]:.2e})"
                            plt.plot(time, a * x * order, linewidth=0.9, label=s)
                            # if (k in dominant ):
                            if k in dominant and k != 0:
                                if name_dominant:
                                    name_dominant = name_dominant + "\n+ " + s
                                else:
                                    name_dominant = s
                else:  # pltoly
                    if self.D.data_df[index_data].index[k] in top_x:
                        # plot only real
                        if k == 0 or k == len(samples) / 2:
                            a = 1
                        else:
                            a = 2
                        if round(freq[k], 1) > 0 and amp[k] < 100:
                            s = f"{a / length * order*amp[k]:.1f}*cos(2\u03C0*{freq[k]:.2f}*t{self.D.data_df[index_data]['phi'].values[k]:+.2f})"
                        else:
                            s = f"{a / length * order*amp[k]:.1e}*cos(2\u03C0*{freq[k]:.2e}*t{self.D.data_df[index_data]['phi'].values[k]:+.2f})"
                        ## For the paper
                        # s = f"{a / length * order*amp[k]:.0f}*cos(2\u03C0*{freq[k]:.2f}*t{self.D.data_df[index_data]['phi'].values[k]:+.2f})"
                        f[-1].add_trace(
                            Scatter(
                                x=time,
                                y=a * x * order,
                                mode="lines",
                                name=s,
                                hovertemplate="<b>Time</b>: %{x:.2f} s"
                                + "<br><b>Amplitude</b>: %{y}"
                                + "<br><b>T</b>: %{text} s",
                                text=len(samples)
                                * [f"{self.D.data_df[index_data]['T'].values[k]:.2f}"],
                                marker_color=(
                                    colors[color_counter]
                                    if color_counter != 1
                                    else "rgb(70,220,70)"
                                ),
                            )
                        )
                        color_counter += 1

                        if k in dominant and k != 0:
                            if name_dominant:
                                name_dominant = name_dominant + " + " + s
                            else:
                                name_dominant = s
                        # in reality it is k/N instead of t
                        # if (k == 0 or k == len(samples)/2):
                        #     s = "%.1ecos(2pi*%.2f*t%+.2f)" % (amp[k], freq[k], self.D.data_df[index_data]["phi"].values[k])
                        #     f[-1].add_trace(Scatter(x=time,y=x,mode='lines',name=s, hovertemplate ='<b>Time</b>: %{x:.2f} s'+ '<br><b>Amplitude</b>: %{y}'+'<br><b>T</b>: %{text} s',text = len(samples)*['%.2f'%(self.D.data_df[index_data]["T"].values[k])]))
                        # else:
                        #     s = "%.1ecos(2pi*%.2f*t%+.2f)" % (a*amp[k], freq[k], self.D.data_df[index_data]["phi"].values[k])
                        #     f[-1].add_trace(Scatter(x=time,y=a*x,mode='lines',name=s, hovertemplate ='<b>Time</b>: %{x:.2f} s'+ '<br><b>Amplitude</b>: %{y}'+'<br><b>T</b>: %{text} s',text = len(samples)*['%.2f'%(self.D.data_df[index_data]["T"].values[k])]))
            if self.dtw:
                threads = []
                print("    '-> \033[1;35mCalculating DTW\033[1;0m")
                threads.append(
                    Thread(
                        target=evaluate_dtw,
                        args=(
                            dominant_X1,
                            sum_all_components,
                            self.D.data_df.iloc[dominant_k1]["freq"],
                        ),
                    )
                )
                threads.append(
                    Thread(
                        target=evaluate_dtw,
                        args=(
                            dominant_X2,
                            sum_all_components,
                            self.D.data_df.iloc[dominant_k2]["freq"],
                        ),
                    )
                )
                # t.append(Thread(target=evaluate_dtw, args=(dominant_X3,sum,self.D.data_df.iloc[dominant_k3]['freq'])))
                for thread in threads:
                    thread.start()

            if "mat" in self.plot_engine:
                plt.plot(
                    time,
                    sum_all_components * order,
                    linestyle="-",
                    marker=".",
                    color="deepskyblue",
                    label="Reconstructed signal",
                )
                plt.xlim(time[0], time[-1])
                plt.ticklabel_format(axis="y", style="sci", scilimits=(-5, 3))
                plt.ticklabel_format(axis="x", style="sci", scilimits=(-5, 3))
                plt.xticks(fontsize=12)
                plt.yticks(fontsize=12)
                plt.ylabel(f"Bandwidth ({unit})", fontsize=17)
                plt.xlabel("Time (s)", fontsize=17)
                plt.grid(True)
                plt.legend(loc="upper left", ncol=2, fontsize=13)
                plt.tight_layout()
                f1.show()

                f2 = plt.figure(figsize=(10, 4))
                if not self.D.original_df.empty:
                    plt.fill_between(
                        self.D.original_df[index_original]["t"],
                        0,
                        self.D.original_df[index_original]["b"] * order,
                        label="Original signal",
                        alpha=0.5,
                        step="post",
                        color="royalblue",
                    )
                    plt.plot(
                        self.D.original_df[index_original]["t"],
                        self.D.original_df[index_original]["b"] * order,
                        drawstyle="steps-post",
                        color="royalblue",
                    )
                # plt.fill_between(time,0,sum,label="Reconstructed signal",alpha=0.5,step="post")
                # plt.plot(time,sum, drawstyle="steps-post")
                plt.fill_between(
                    time,
                    0,
                    self.D.data_df[index_data]["b_sampled"] * order,
                    label="Discrete signal",
                    alpha=0.6,
                    step="post",
                    color="red",
                )
                plt.plot(
                    time,
                    self.D.data_df[index_data]["b_sampled"] * order,
                    drawstyle="steps-post",
                    color="red",
                )
                if found:
                    plt.fill_between(
                        time,
                        0,
                        y2=sum_dominant * order,
                        label=name_dominant,
                        alpha=0.6,
                        step="post",
                        color="limegreen",
                    )
                    plt.plot(
                        time,
                        sum_dominant * order,
                        drawstyle="steps-post",
                        color="limegreen",
                    )
                if self.recon:
                    plt.fill_between(
                        time,
                        0,
                        y2=sum_top_2 * order,
                        label="Recon. top 2 signal",
                        alpha=0.6,
                        step="post",
                        color="gold",
                    )
                    plt.plot(
                        time, sum_top_2 * order, drawstyle="steps-post", color="gold"
                    )
                    plt.fill_between(
                        time,
                        0,
                        y2=sum_top_3 * order,
                        label="Recon. top 3 signal",
                        alpha=0.6,
                        step="post",
                        color="cyan",
                    )
                    plt.plot(
                        time, sum_top_3 * order, drawstyle="steps-post", color="cyan"
                    )
                    plt.fill_between(
                        time,
                        0,
                        y2=sum_top_5 * order,
                        label="Recon. top 5 signal",
                        alpha=0.6,
                        step="post",
                        color="purple",
                    )
                    plt.plot(
                        time, sum_top_5 * order, drawstyle="steps-post", color="purple"
                    )
                    plt.fill_between(
                        time,
                        0,
                        y2=sum_top_10 * order,
                        label="Recon. top 10 signal",
                        alpha=0.6,
                        step="post",
                        color="slategrey",
                    )
                    plt.plot(
                        time,
                        sum_top_10 * order,
                        drawstyle="steps-post",
                        color="slategrey",
                    )
                plt.xlim(time[0], time[-1])
                plt.ticklabel_format(axis="y", style="sci", scilimits=(-5, 3))
                plt.ticklabel_format(axis="x", style="sci", scilimits=(-5, 3))
                plt.xticks(fontsize=12)
                plt.yticks(fontsize=12)
                plt.ylabel(f"Bandwidth ({unit})", fontsize=17)
                plt.xlabel("Time (s)", fontsize=17)
                plt.grid(True)
                plt.legend(loc="upper left", ncol=1, fontsize=13)
                plt.tight_layout()
                f2.show()

                f3 = plt.figure(figsize=(10, 4))
                # settings
                full = False
                percent = True

                if not full:
                    n_samples = len(freq)
                    last = int(n_samples / 2) - 1
                    amp = 2 * amp[0 : last + 1]
                    amp[last] = amp[last] / 2
                    amp[0] = amp[0] / 2
                    freq = freq[0 : int(n_samples / 2)]

                mode = "Amplitude"
                if self.psd:
                    amp = amp * amp / len(amp)
                    mode = "Power"

                if percent:
                    mode = "Normed " + mode + " (%)"
                    amp = 100 * amp / amp.sum()

                plt.bar(
                    freq,
                    amp,
                    edgecolor="k",
                    width=freq[1] - freq[0],
                )
                plt.ticklabel_format(axis="y", style="sci", scilimits=(-5, 3))
                plt.ticklabel_format(axis="x", style="sci", scilimits=(-5, 3))
                plt.xticks(fontsize=12)
                plt.yticks(fontsize=12)
                plt.ylabel(mode, fontsize=17)
                plt.xlabel("Frequency (Hz)", fontsize=17)
                plt.grid(True)
                plt.legend(loc="upper left", ncol=2, fontsize=13)
                plt.tight_layout()
                f3.show()

            else:  # pltoly
                #! Reconstructed plot
                #!######################
                colors = px.colors.qualitative.Plotly

                f[-1].add_trace(
                    Scatter(
                        x=time,
                        y=sum_all_components * order,
                        mode="lines+markers",
                        name="Discrete signal",
                        line={"shape": "hv"},
                        marker_color="rgb(180,30,30)",
                    )
                )
                f[-1].update_layout(
                    xaxis_title="Time (s)",
                    yaxis_title=f"Bandwidth ({unit})",
                    width=width,
                    height=height / 1.1,
                    template=template,
                )
                f[-1].update_layout(
                    legend=dict(
                        orientation="h", yanchor="top", y=0.99, xanchor="left", x=0.01
                    )
                )
                # f[-1].update_layout(legend=dict(orientation="h", yanchor="bottom",y=1.02, xanchor="right", x=1))
                f[-1].update_xaxes(range=[time[0], time[-1]])
                f[-1] = format_plot(f[-1])
                f[-1].show(config=conf)

                f.append(go.Figure())
                fill = "tozeroy"
                # if len(time) > 1e3:
                #     fill = None
                #     print ('too many points, removing fill')

                #! Dominant plot
                #!######################
                if not self.D.original_df.empty and (self.render == "dynamic" or  len(time) < 1e3):
                    f[-1].add_trace(
                        Scatter(
                            x=self.D.original_df[index_original]["t"],
                            y=self.D.original_df[index_original]["b"] * order,
                            mode="lines+markers",
                            name="Original signal",
                            fill=fill,
                            line={"shape": "hv"},
                            marker_color="rgb(0,150,250)",
                        )
                    )
                else:
                    console.print(
                        "[orange]Too many data points for original signal[/]\n"
                        "[orange]>> skipping plotting it[/]\n"
                        "[orange]>> use ftio ... -e mat to show this plot or set ftio ... -re static[/]\n"
                    )
                # f[-1].add_trace(Scatter(x=time,y=sum,mode='lines+markers',name="Reconstructed signal",fill=fill, line={"shape": 'hv'},visible=visible))
                f[-1].add_trace(
                    Scatter(
                        x=time,
                        y=self.D.data_df[index_data]["b_sampled"] * order,
                        mode="lines+markers",
                        name="Discrete signal",
                        fill=fill,
                        line={"shape": "hv"},
                        marker_color="rgb(180,30,30)",
                    )
                )
                if found:
                    if r > 1000:
                        fill = None
                    f[-1].add_trace(
                        Scatter(
                            x=time,
                            y=sum_dominant * order,
                            mode="lines+markers",
                            name=name_dominant,
                            fill=fill,
                            line={"shape": "hv"},
                            marker_color="rgb(70,220,70)",
                        )
                    )
                if self.recon:
                    f[-1].add_trace(
                        Scatter(
                            x=time,
                            y=sum_top_2 * order,
                            mode="lines+markers",
                            name="Recon. top 2 signal",
                            fill=fill,
                            line={"shape": "hv"},
                            visible="legendonly",
                        )
                    )
                    f[-1].add_trace(
                        Scatter(
                            x=time,
                            y=sum_top_3 * order,
                            mode="lines+markers",
                            name="Recon. top 3 signal",
                            fill=fill,
                            line={"shape": "hv"},
                            visible="legendonly",
                        )
                    )
                    f[-1].add_trace(
                        Scatter(
                            x=time,
                            y=sum_top_5 * order,
                            mode="lines+markers",
                            name="Recon. top 5 signal",
                            fill=fill,
                            line={"shape": "hv"},
                            visible="legendonly",
                        )
                    )
                    f[-1].add_trace(
                        Scatter(
                            x=time,
                            y=sum_top_10 * order,
                            mode="lines+markers",
                            name="Recon. top 10 signal",
                            fill=fill,
                            line={"shape": "hv"},
                            visible="legendonly",
                        )
                    )
                rangeslider(
                    f[-1],
                    time,
                    2 / self.D.data_df.iloc[dominant_k1]["freq"],
                    slider_cond,
                )
                f[-1].update_layout(
                    xaxis_title="Time (s)",
                    yaxis_title=f"Bandwidth ({unit})",
                    font=font_settings,
                    # width=1.05*width,
                    width=width,
                    height=height / 1.1,
                    # title="Time Plot (Ranks %i)" % r,
                    template=template,
                )
                f[-1].update_layout(
                    legend=dict(yanchor="top", y=0.99, xanchor="left", x=0.01)
                    # legend=dict(yanchor="top", y=0.99, xanchor="right", x=.99)
                )

                f[-1].update_xaxes(range=[time[0], time[-1]])
                if isinstance(sum_dominant, list):
                    f[-1].update_yaxes(
                        range=[
                            min(sum_dominant),
                            1.2 * self.D.original_df[index_original]["b"].max(),
                        ]
                    )
                # f[-1].update_yaxes(range=[-2000, 72000])
                f[-1] = format_plot(f[-1])
                # f[-1].show(config=conf)

                #! Frequency spectrum
                #!######################
                fig_tmp = plot_one_spectrum(self.psd, freq, amp, False, True)
                fig_tmp.update_layout(
                    font=font_settings,
                    width=width,
                    height=height / 1.3,
                    coloraxis_colorbar=dict(
                        yanchor="top", y=1, x=1, ticks="outside", title=""
                    ),
                    template=template,
                    # title="Spectrum (Ranks %i)" % r
                )
                # fig_tmp.show(config=conf)
                f.append(fig_tmp)

            if self.dtw:
                for thread in threads:
                    thread.join()

            # f.append(go.Figure())
            # f[-1].add_trace(go.Bar(x=samples,y=amp,marker_color='rgb(26, 118, 255)',marker_line=dict(width=1, color=color_bar), hovertemplate ='<br><b>k</b>: %{x}<br>' + '<b>Amplitude</b>: %{y:.2e}' +'<br><b>Frequency</b>: %{text} Hz<br>',text = ['%.2f'%i for i in freq]))
            # rangeslider(f[-1],samples,int(len(samples)/5),len(samples) > 1e3 and True)
            # f[-1].update_layout(xaxis_title='Frequency bins', yaxis_title='Amplitude',font=font_settings, width=width, height=height, title = 'Frequency Plot (Ranks %i)'%r, coloraxis_colorbar=dict(yanchor="top", y=1, x=0,ticks="outside"), template = template)

            if "plotly" == self.plot_engine:
                fig_tmp = plot_one_spectrum(self.psd, freq, amp, True)
                fig_tmp.update_traces(marker_line=dict(width=0.1, color=color_bar))
                fig_tmp.update_layout(
                    font=font_settings,
                    width=width,
                    height=height / 1.3,
                    # title="Full Spectrum (Ranks %i)" % r,
                    coloraxis_colorbar=dict(
                        yanchor="top", y=1, x=1, ticks="outside", title=""
                    ),
                    template=template,
                )
                f.append(fig_tmp)
                rangeslider(
                    f[-1],
                    freq,
                    freq[int(len(freq) / 5)],
                    len(samples) > 1e3 and slider_cond,
                )

                # bar_plot.update_layout(barmode="group", template=template)
                # bar_plot.add_trace(
                #     go.Bar(
                #         x=samples,
                #         y=amp,
                #         textposition="inside",
                #         textangle=0,
                #         name=("Ranks %i") % r,
                #         hovertemplate="<br><b>Frequency</b>: %{x} Hz<br>"
                #         + "<b>Amplitude</b>: %{y:.2e}"
                #         + "<br><b>T</b>: %{text} s<br>",
                #         text=["%.2f" % i for i in self.D.data_df[index_data]["T"]],
                #     )
                # )
                # bar_plot.update_layout(
                #     xaxis_title="Frequency bins",
                #     yaxis_title="Amplitude",
                #     width=width,
                #     height=height,
                #     title="Frequency Plot",
                #     font=font_settings,
                #     barmode="group",
                # )

            #! 3d plot
            # prediction_plot.add_traces(list(px.scatter_3d(self.D.data_df.iloc[top_3], x='A', y='ranks', z='k' ,color='T', symbol='ranks',opacity=0.7,hover_data={'ranks':True, 'k'  :True, 'A':':.2f', 'F': (':.2f (Hz)',self.D.data_df.iloc[top_3]['freq']), 'T': ':.2f (s)'}).select_traces()))

        if "plotly" == self.plot_engine:
            # bar_plot.update_xaxes(range=[0, 100])
            # f.append(bar_plot)

            #! 3d plot
            # prediction_plot.update_layout(template = template,margin=dict(l=0, r=0, b=0, t=0), width=width, height = height, title = 'Frequency Prediction', font=font_settings, coloraxis_colorbar=dict(title='Perodicity (s)', yanchor="top", y=1, x=0,ticks="outside"),scene = dict(xaxis_title='Amplitude',yaxis_title='Ranks',zaxis_title='Frequency (Hz)',xaxis=dict(showgrid=True, gridwidth=1, gridcolor='gray'),yaxis=dict(showgrid=True, gridwidth=1, gridcolor='gray'),zaxis=dict(showgrid=True, gridwidth=1, gridcolor='gray')))
            # prediction_plot.layout.scene.aspectratio = {'x':2, 'y':1, 'conf':1}
            # f.append(prediction_plot)

            # fig_tmp = px.scatter_3d(self.D.data_df, x='k', y='ranks', z='A',color='A', symbol='ranks', opacity=0.7)
            # tight layout
            # fig_tmp.update_layout(
            # template = template,
            # margin=dict(l=0, r=0, b=0, t=0), width=width, height = height, title = 'Frequency Plot',
            # font=font_settings,
            # coloraxis_colorbar=dict(title='Frequency (Hz)', yanchor="top", y=1, x=0, ticks="outside"),
            # scene = dict(xaxis_title='Frequency (Hz)',yaxis_title='Ranks', zaxis_title='Bandwidth (B/s)',
            # xaxis=dict(showgrid=True, gridwidth=1, gridcolor='gray'),
            # yaxis=dict(showgrid=True, gridwidth=1, gridcolor='gray'),
            # zaxis=dict(showgrid=True, gridwidth=1, gridcolor='gray')))
            # fig_tmp.layout.scene.aspectratio = {'x':2, 'y':1, 'conf':1}
            # f.append(fig_tmp)

            # f.append(go.Figure())
            # f[-1].add_traces(
            #     list(
            #         px.line(
            #             self.D.data_df.iloc[top],
            #             x="ranks",
            #             y="T",
            #             color="k",
            #             markers=True,
            #             hover_data={
            #                 "ranks": True,
            #                 "k": True,
            #                 "A": ":.2f",
            #                 "F": (":.2f (Hz)", self.D.data_df.iloc[top]["freq"]),
            #                 "T": True,
            #             },
            #         ).select_traces()
            #     )
            # )
            # f[-1].update_layout(
            #     xaxis_title="Ranks",
            #     yaxis_title="Period (s)",
            #     font=font_settings,
            #     width=width,
            #     height=height,
            #     title="Prediction",
            #     template=template,
            # )
            create_html(f, self.render, conf, "freq")
        else:
            input()


##Fill DTW Matrix
def fill_dtw_cost_matrix(s1, s2):
    l_s_1, l_s_2 = len(s1), len(s2)
    cost_matrix = np.zeros((l_s_1 + 1, l_s_2 + 1))
    for i in range(l_s_1 + 1):
        for j in range(l_s_2 + 1):
            cost_matrix[i, j] = np.inf
    cost_matrix[0, 0] = 0

    for i in range(1, l_s_1 + 1):
        for j in range(1, l_s_2 + 1):
            cost = abs(s1[i - 1] - s2[j - 1])
            # take last min from the window
            prev_min = np.min(
                [
                    cost_matrix[i - 1, j],
                    cost_matrix[i, j - 1],
                    cost_matrix[i - 1, j - 1],
                ]
            )
            cost_matrix[i, j] = cost + prev_min

    return cost_matrix[-1, -1]


##Call DTW function


def fdtw(s1, s2):
    distance, path = fastdtw(s1, s2, dist=euclidean)
    return distance, path


def rangeslider(f, arr, limit, cond="", point_limit=2.5e3):
    if isinstance(cond, str):
        cond = len(arr) > point_limit
    try:
        if cond:
            if limit < arr[-1]:
                myrange = [arr[0], limit]
                f.update_layout(
                    xaxis=dict(
                        rangeselector=dict(
                            buttons=[
                                dict(
                                    count=2,
                                    label="phase",
                                    step="month",
                                    stepmode="backward",
                                ),
                                dict(label="all", step="all"),
                            ]
                        ),
                        rangeslider=dict(visible=True, range=[arr[0], arr[-1]]),
                        type="linear",
                        range=myrange,
                    )
                )  # change type to date to show buttons
            else:
                pass
        else:
            pass
    except:
        try:
            myrange = [arr[0], arr[int(len(arr) / 5)]]
            f.update_layout(
                xaxis=dict(
                    rangeselector=dict(
                        buttons=[
                            dict(
                                count=2,
                                label="phase",
                                step="month",
                                stepmode="backward",
                            ),
                            dict(label="all", step="all"),
                        ]
                    ),
                    rangeslider=dict(visible=True, range=[arr[0], arr[len(arr) - 1]]),
                    type="linear",
                    range=myrange,
                )
            )  # change type to date to show buttons
        except:
            pass


def evaluate_dtw(discret_arr, original_discret_signal, freq):
    dtw_k1, _ = fastdtw(discret_arr, original_discret_signal, dist=euclidean)
    print("    '-> \033[1;32mfreq %.2f Hz\033[1;0m --> dtw: %d" % (freq, dtw_k1))


def plot_spectrum(
    amp: np.ndarray, freq: np.ndarray, mode: str = "Amplitude", percent: bool = False
):
    template = "plotly"
    name = "Amplitude" if "amp" in mode.lower() else "Power"
    unit = ""
    if percent:
        name = "Normed " + name
        unit = unit + " (%)"
        amp = 100 * amp / amp.sum()

    df_tmp = pd.DataFrame(
        data={
            "A": amp,
            "freq": freq,
        }
    )
    # fig_tmp = px.bar(df_tmp, x="freq", y="A", color="A", color_continuous_scale='Bluered')
    # fig_tmp = px.bar(df_tmp, x="freq", y="A", color="A", color_continuous_scale=["rgb(0,10,130)", "rgb(130,30,130)", "rgb(255,50,50)"])
    # fig_tmp = px.bar(df_tmp, x="freq", y="A", color="A", color_continuous_scale="plasma")
    # cyan - purple
    # fig_tmp = px.bar(df_tmp, x="freq", y="A", color="A", color_continuous_scale=[ "rgb(90, 40, 90)", "rgb(113, 221, 255)" ])
    fig_tmp = px.bar(
        df_tmp,
        x="freq",
        y="A",
        color="A",
        color_continuous_scale=["rgb(0,50,150)", "rgb(150,50,150)", "rgb(255,50,0)"],
    )

    fig_tmp.update_traces(
        marker_line=dict(width=0.2, color="black"),
        hovertemplate="<b>Frequency</b>: %{x:.2e} Hz <br><b>"
        + f"{name}"
        + "</b>: %{y:.2e}"
        + f"{unit}",
    )
    fig_tmp.update_layout(
        xaxis_title="Frequency (Hz)",
        yaxis_title=name + unit,
        coloraxis_colorbar=dict(yanchor="top", y=1, x=0, ticks="outside"),
        template=template,
    )
    fig_tmp = format_plot(fig_tmp)
    return fig_tmp, name + unit


def plot_both_spectrums(args, freq: np.ndarray, amp: np.ndarray, full: bool = True):
    fig_1 = make_subplots(
        rows=2,
        cols=2,
        specs=[[{"colspan": 2}, None], [{"colspan": 2}, None]],
    )

    start = 1

    if not full:
        indecies = np.arange(start, int(len(amp) / 2) + 1)
        freq = freq[indecies]
        amp = 2 * amp[indecies]
        amp[-1] = amp[-1] / 2
        if start == 0:
            amp[0] = amp[0] / 2

    if args.psd:
        power = amp
        amplitude = np.sqrt(amp) * len(amp)
    else:
        power = amp * amp / len(amp)
        amplitude = amp

    fig_tmp, name_plt = plot_spectrum(amplitude, freq, "Amplitude", False)
    layout = fig_tmp.layout
    for trace in list(fig_tmp.select_traces()):
        trace.update(marker={"coloraxis": "coloraxis"})
        fig_1.append_trace(trace, row=1, col=1)
    fig_tmp, name_plt1 = plot_spectrum(power, freq, "Power", True)
    for trace in list(fig_tmp.select_traces()):
        trace.update(marker={"coloraxis": "coloraxis2"})
        fig_1.append_trace(trace, row=2, col=1)
    fig_1.update_layout(layout)
    fig_1.update_layout(
        coloraxis={
            "colorbar": {
                "x": 1,
                "len": 0.4,
                "y": 0.8,
            },
        },
        coloraxis2={
            "colorbar": {
                "x": 1,
                "len": 0.4,
                "y": 0.2,
            },
        },
    )

    return fig_1, [name_plt, name_plt1]


def plot_one_spectrum(
    psd_flag: bool,
    freq: np.ndarray,
    amp: np.ndarray,
    full: bool = True,
    percent: bool = False,
) -> go.Figure:
    if not full:
        n_samples = len(freq)
        last = int(n_samples / 2) - 1
        amp = 2 * amp[0 : last + 1]
        amp[last] = amp[last] / 2
        amp[0] = amp[0] / 2
        freq = freq[0 : int(n_samples / 2)]

    if psd_flag:
        amp = amp * amp / len(amp)
        mode = "Power"
    else:
        mode = "Amplitude"

    fig_tmp, _ = plot_spectrum(amp, freq, mode, percent)
    return fig_tmp


def convert_and_plot(data, dfs: list, args) -> None:
    """convert from ftio and plot

    Args:
        data (_type_): _description_
        dfs (list): _description_
        args (argparse): _description_
    """
    freq_plot = FreqPlot(True)
    if any(x in args.engine for x in ["mat", "plot"]):
        freq_plot.add_df(len(data), dfs[0], dfs[1], dfs[2], dfs[3])
    freq_plot.set(
        {
            "render": args.render,
            "engine": args.engine,
            "dtw": args.dtw,
            "recon": args.reconstruction,
            "psd": args.psd,
        }
    )
    freq_plot.Plot()



""" Wavlet functions (continous and discrete) + plot methods
    """
import numpy as np
import pywt
from scipy import signal
import matplotlib.pyplot as plt

# Wavelet
# ----------
def wavelet_cont(b_sampled, wavelet, level, freq):
    # Continouse wavelet:
    n = len(b_sampled)
    if isinstance(level, str) and "auto" in level:
        level = pywt.dwt_max_level(n, wavelet)  # pywt.dwt_max_level(n,wavelet)
    else:
        level = int(level)
    print("    '-> \033[1;32mDecomposition level is %d\033[1;0m" % level)
    # [coefficients, frequencies] = pywt.cwt(b_sampled, np.arange(1,5), 'morl')
    # scale = pywt.frequency2scale('morl', np.array([1,10,100])/freq)
    # scale = pywt.frequency2scale('morl', np.arange(freq/100,freq)/freq)
    # scale = 2**np.arange(0,level) #2** mimcs the DWT
    scale = np.arange(1, level)  # 2** mimcs the DWT

    check_wavelet(wavelet, "continuous")
    print(pywt.scale2frequency(wavelet, scale) / (1 / freq))

    [coefficients, frequencies] = pywt.cwt(
        b_sampled, scale, wavelet, sampling_period=1 / freq
    )
    return coefficients, frequencies, scale


def plot_wave_cont(b_sampled, frequencies, freq, scale, t, coefficients):
    time_disc = t[0] + 1 / freq * np.arange(0, len(b_sampled))
    power = (
        abs(coefficients)
    ) ** 2  # probably error on https://ataspinar.com/2018/12/21/a-guide-for-using-the-wavelet-transform-in-machine-learning/
    counter = 0
    for (
        i
    ) in (
        scale
    ):  # see Continuous wavelet transform properties @ https://en.wikipedia.org/wiki/Continuous_wavelet_transform
        power[counter] = 1 / i * power[counter]
        counter = counter + 1
    fig, ax = plt.subplots(figsize=(12, 4))
    im = ax.contourf(
        time_disc,
        frequencies,
        power,
        np.arange(0, power.max(), power.max() / 10),
        extend="neither",
        cmap=plt.cm.seismic,
    )
    # fig.colorbar(im, cax=cbar_ax, orientation="vertical")
    fig.colorbar(im, ax=ax)
    ax.set_ylabel("Frequency (Hz)", fontsize=18)
    ax.set_xlabel("Time (s)", fontsize=18)
    plt.xticks(fontsize=16)
    plt.yticks(fontsize=16)
    fig.tight_layout()
    plt.show()
    return fig


def wavelet_disc(b_sampled, wavelet, level):
    # Discrete Wavelet:
    check_wavelet(wavelet, "discrete")
    if isinstance(level, str) and "auto" in level:
        level = pywt.dwt_max_level(
            len(b_sampled), wavelet
        )  # pywt.dwt_max_level(n,wavelet)
    else:
        level = int(level)
    print("    '-> \033[1;32mDecomposition level is %d\033[1;0m" % level)
    coffs = pywt.wavedec(b_sampled, wavelet, level=level)
    return coffs


def plot_wave_disc(b_sampled, coffs, t, freq, level, wavelet, b):
    n = len(b_sampled)
    time_disc = t[0] + 1 / freq * np.arange(0, n)
    reconstructed_signal = pywt.waverec(coffs, wavelet, "smooth")
    f = []
    f1 = plt.figure(figsize=(12, 4))
    plt.plot(t, b, label="signal")
    plt.plot(
        time_disc,
        reconstructed_signal[0 : len(time_disc)],
        label="reconstructed levels %d" % level,
        linestyle="--",
    )
    # plt.plot(time,reconstructed_signal[0:len(time)], label='reconstructed levels %d', linestyle='--')
    plt.legend(loc="upper right")
    plt.title("single reconstruction", fontsize=20)
    plt.xlabel("time axis", fontsize=16)
    plt.ylabel("Amplitude", fontsize=16)

    # ? use recon or coffs to either create the paritallly reconstructed signal or the cofficent of the DTW:
    # ? recon -> idea is that all values are extracted just before reconstructing the signal. This adds less resolution to the upper frequencies
    # ? x ---->.g -->↓ 2 --> C1 --> ↑2 -->* g-1  #   captuere   # -->+ --> x
    # ?     '->.h -->↓ 2 --> C2 --> ↑2 -->* h-1  #      here    #  --^
    use = "recon"
    # use = 'coffs' #hold the signal, not upsample!
    cc = np.zeros((level + 1, n))
    for i in np.arange(start=level, stop=0, step=-1):
        # coffs ->  [cA_n, cD_n, cD_n-1, …, cD2, cD1]
        # reconstruction by 1) upsampling and 2) multiplying with the appropiate filter
        # page 29 on https://www.corsi.univr.it/documenti/OccorrenzaIns/matdid/matdid358630.pdf
        # GoH0+G1H1 = 1 -> Multiply a with H1 and d with G1
        # https://medium.com/@shouke.wei/process-of-discrete-wavelet-transform-iii-wavelet-partial-reconstruction-ca7a8f9420dc
        if "recon" in use:
            cc[level - i + 1] = pywt.upcoef(
                "d", coffs[level - i + 1], wavelet, level=i
            )[:n]
        else:
            # Wrong: only upsample: ‘↑ 2’ denotes ‘upsample by 2’ (put 0’s before values)
            # cc[level-i+1,::2**(i)] = coffs[level-i+1]
            # Correct: upsample is used for reconstruction, we only want to visualize -> hold the data constatnt during sample step

            counter = 0
            for j in range(0, n):
                if j % 2**i == 0:
                    cc[level - i + 1, j] = coffs[level - i + 1][counter]
                    counter = counter + 1
                else:
                    cc[level - i + 1, j] = cc[level - i + 1, j - 1]

    if "recon" in use:
        cc[0] = pywt.upcoef("a", coffs[0], wavelet, level=level)[:n]
        fig, ax = plt.subplots(nrows=level + 2, ncols=1, figsize=(12, 12))
        ax[0].plot(t, b, label="signal")
        for i in range(0, level + 1):
            if i == 0:
                sum = cc[0]
            else:
                sum = sum + cc[i]
        ax[0].plot(time_disc, sum, label="sum")
        for i in np.arange(start=level, stop=0, step=-1):
            ax[level - i + 2].plot(time_disc, cc[level - i + 1])
            ax[level - i + 2].set_title(
                f"reconstruction at level {i} from detailed coff -> [{freq / 2 ** (i + 1)}, {freq / 2**i}] Hz"
            )
        ax[1].plot(time_disc, cc[0])
        ax[1].set_title(
            f"reconstruction at level {level} from approximated coff -> [0, {freq / 2 ** (level + 1)}] Hz"
        )
        fig.legend()
        fig.tight_layout()
        f.append(fig)

    else:
        counter = 0
        for j in range(0, n):
            if j % 2**level == 0:
                cc[0, j] = coffs[0][counter]
                counter = counter + 1
            else:
                cc[0, j] = cc[0, j - 1]

    # Use freq or level to plot either the frequency or the level in the y-axis
    show = "freq"
    f_2 = plt.figure(figsize=(12, 4))
    plt.xlabel("Time (s)", fontsize=18)
    # plt.title('Discrete Wavelet Transform with %d decompositions'%level)
    if "freq" in show:
        plt.ylabel("Frequency (Hz)", fontsize=18)
        plt.xticks(fontsize=18)
        plt.yticks(fontsize=18)
        y = np.concatenate(
            [np.array([0]), freq / 2 ** np.arange(start=level + 1, stop=0, step=-1)]
        )
        x = (
            -2 / freq + t[0] + 1 / freq * np.arange(0, len(b_sampled) + 1)
        )  # ? add corner shifted by half a sample step
        X, Y = np.meshgrid(x, y)
        # plt.pcolormesh(X, Y,abs(cc),cmap=plt.cm.coolwarm,shading='flat')
        plt.pcolormesh(X, Y, abs(cc), cmap=plt.cm.seismic, shading="flat")
    else:
        y = np.arange(start=level, stop=-1, step=-1)
        x = t[0] + 1 / freq * np.arange(0, len(b_sampled))
        X, Y = np.meshgrid(x, y)
        plt.pcolormesh(X, Y, abs(cc), cmap=plt.cm.coolwarm)
        plt.ylabel("decomposition level")
        # cc = np.flip(cc,0)
        plt.gca().invert_yaxis()

    # plt.pcolormesh(X, Y, cc,cmap=plt.cm.seismic)
    # plt.pcolormesh(X, Y,abs(cc),cmap=plt.cm.seismic,shading='flat')
    # plt.plot(X.flat, Y.flat, 'x', color='m')
    plt.colorbar()
    f.append(f_2)
    plt.tight_layout()
    f.append(f1)
    return cc, f


def check_wavelet(wavelet, mode="discrete"):
    """check that the wavelet selected is supported

    Args:
        wavelet (str): wavelet name
        mode (str, optional): Wavelet mode. Defaults to "discrete".

    Raises:
        Exception: exits the program on unsupported mode
    """
    check = False
    for supported_wavelet in pywt.wavelist(kind=mode):
        if wavelet == supported_wavelet:
            check = True
            break
    if check is False:
        raise ValueError(
            "Unsupported wavelet specified, suported modes are: %s\nsee: https://pywavelets.readthedocs.io/en/latest/ref/wavelets.html "
            % pywt.wavelist(kind=mode)
        )
        # sys.exit()
    else:
        return


def welch(b, freq):
    """Welch method for spectral density estimation.

    Args:
        b (list[float]): bandwidth over time
        freq (float): sampling frequency
    """
    # f, Pxx_den = signal.welch(b, freq, 'flattop', 10000, scaling='spectrum', average='mean')
    f, Pxx_den = signal.welch(
        b, freq, "flattop", freq * 256, scaling="spectrum", average="mean"
    )
    plt.semilogy(f, np.sqrt(Pxx_den))
    plt.xlabel("frequency [Hz]")
    # plt.ylabel('PSD [V**2/Hz]')
    plt.ylabel("Linear spectrum [V RMS]")
    plt.show()


"""
------------------------------
Outlier detection methods
------------------------------
"""
# DB-scan
from __future__ import annotations
from sklearn.cluster import DBSCAN
from sklearn.neighbors import NearestNeighbors
from kneed import KneeLocator
from rich.panel import Panel
# Isolation forest
from sklearn.ensemble import IsolationForest
# Lof
from sklearn.neighbors import LocalOutlierFactor
# find_peaks
from scipy.signal import find_peaks
# all
import numpy as np
from ftio.freq.anomaly_plot import  plot_outliers,  plot_decision_boundaries


def outlier_detection(amp:np.ndarray, freq_arr:np.ndarray, args) -> tuple[list[float], np.ndarray, Panel]:
    """Find the outliers in the samples

    Args:
        A (list[float]): Amplitudes array
        freq_arr (list[float]): frequency array
        args (object, optional): arguments containgin: outlier detection method (Z-Score, DB-Scan). 
        Defaults to 'Z-score'.

    Returns:
        dominant_index (list[float]): indecies of dominant frequencies
        conf (list[float]): confidence in the predictions
        Panel: text to display using rich
    """
    methode = args.outlier
    text = ""    
    if methode.lower() in ["z-score", "zscore"]:
        dominant_index, conf, text = z_score(amp, freq_arr, args)
        title = "Z-score"
    elif methode.lower() in ["dbscan", "db-scan", "db"]:
        dominant_index, conf, text = db_scan(amp, freq_arr, args)
        title = "DB-Scan"
    elif methode.lower() in ["isolation_forest", "forest"]:
        dominant_index, conf, text = isolation_forest(amp, freq_arr, args)
        title = "Isolation Forest"
    elif methode.lower() in ["local outlier factor", "lof"]:
        dominant_index, conf, text = lof(amp, freq_arr, args)
        title = "Local Outlier Factor"
    elif methode.lower() in ["find peaks", "peaks", "peak"]:
        dominant_index, conf, text = peaks(amp, freq_arr, args)
        title = "Find Peaks"
    else:
        dominant_index, conf = [],np.array([])
        raise NotImplementedError("Unsupported method selected")
    # sort dominant index according to confidence
    # if len(dominant_index) > 1:
    #     print(f"conf: {conf} and dominant_index {dominant_index}")
    #     tmp = conf[dominant_index]
    #     dominant_index = np.array(dominant_index)
    #     dominant_index = list(dominant_index[np.argsort(tmp)])
    text = Panel.fit(text[:-1], style="white", border_style='green', title=title, title_align='left')

    return dominant_index, conf, text


# ?#################################
# ? Z-score
# ?#################################
def z_score(
    amp: np.ndarray, freq_arr: np.ndarray, args
) -> tuple[list[float], np.ndarray, str]:
    """calculates the outliers using zscore

    Args:
        amp (np.ndarray): amplitude or psd
        freq_arr (np.ndarray): freuqencies
        args (argsparse): arguments

    Returns:
        tuple[list[float], np.ndarray, str]: [dominant freuqency/ies, confidence, text]
    """
    text = "[green]Spectrum[/]: Amplitude spectrum\n"
    if args.psd:
        amp = amp*amp/len(amp)
        text = "[green]Spectrum[/]: Power spectrum\n"

    indecies = np.arange(1, int(len(amp) / 2) + 1)
    amp_tmp = np.array(2 * amp[indecies])
    # norm the data
    amp_tmp = amp_tmp/ amp_tmp.sum()
    
    tol = args.tol
    dominant_index = []
    mean = np.mean(amp_tmp)
    std = np.std(amp_tmp)
    z_k = abs(amp_tmp - mean) / std
    conf = np.zeros(len(z_k))
    # find outliers
    index = np.where((z_k/np.max(z_k) > tol) & (z_k > 3))
    text += f"[green]mean[/]: {mean/np.sum(amp_tmp):.3e}\n[green]std[/]: {std:.3e}\n"
    text += f"Frequencies with Z-score > 3 -> [green]{len(z_k[z_k>3])}[/] candidates\n"
    text += f"         + Z > Z_max*{tol*100}% > 3 -> [green]{len(index[0])}[/] candidates\n"
    # text += f"         + Z > Z_max*{tol*100}% > 3 -> [green]{len(z_k[(z_k>np.max(z_k)*tol) & (z_k>3)])}[/] candidates\n"
    index, removed_index, msg = remove_harmonics(freq_arr, amp_tmp,  indecies[index[0]])
    text+= msg
    
    if len(index) == 0:
        text += "[red]No dominant frequency -> Signal might be not perodic[/]\n"
    else:
        removed_index = [i-1 for i in removed_index] #tmp starts at 1
        # conf[index] = (z_k[index]/max_z  + z_k[index]/np.sum(z_k[index]) + 1/np.sum(z_k > 3))/3
        # calculate the confidence:
        # 1) check z_k/max_zk > tol
        tmp = z_k/np.max(z_k) > tol
        tmp[removed_index] = False
        conf[tmp] += z_k[tmp]/np.sum(z_k[tmp])
        
        # 2) check z_k > 0
        tmp = z_k > 3
        tmp[removed_index] = False
        conf[tmp] += z_k[tmp]/np.sum(z_k[tmp])
        conf = conf/2
        conf = np.array(conf)
        

        # get dominant index
        dominant_index, msg = dominant(index, freq_arr, conf)
        text+= msg
        
    if "plotly" in args.engine:
        i = np.repeat(1, len(indecies))
        if len(dominant_index) != 0:
            i[np.array(dominant_index) - 1] = -1
        plot_outliers(args,freq_arr, amp, indecies, conf, i)

    return dominant_index, conf, text


# ?#################################
# ? DB-Scan
# ?#################################
def db_scan(
    amp: np.ndarray, freq_arr: np.ndarray, args
) -> tuple[list[float], np.ndarray, str]:
    """calculates the outliers using dbscan

    Args:
        amp (np.ndarray): amplitude or psd
        freq_arr (np.ndarray): freuqencies
        args (argsparse): arguments

    Returns:
        tuple[list[float], np.ndarray, str]: [dominant freuqency/ies, confidence]
    """
    text = "[green]Spectrum[/]: Amplitude spectrum\n" 
    if args.psd:
        amp = amp*amp/len(amp)
        text = "[green]Spectrum[/]: Power spectrum\n"
        
    indecies = np.arange(1, int(len(amp) / 2) + 1)
    amp_tmp = np.array(2 * amp[indecies])
    freq_arr_tmp = np.array(freq_arr[indecies])
    min_pts = 2

    # norm the data
    # d = np.vstack((freq_arr_tmp / freq_arr_tmp.max(), amp_tmp / amp_tmp.max())).T
    #! norm over sum for amplitude with power spectrum 
    d = np.vstack((freq_arr_tmp / freq_arr_tmp.max(), amp_tmp / amp_tmp.sum())).T

    eps_mode = "range"
    if eps_mode == "avr":
        text += "Calculating eps using average\n"
        try:
            eps = np.sqrt(pow(d[:, 1].mean(), 2) + pow(d[1, 0] - d[0, 0], 2))
        except ValueError:
            eps = 0.1
        conf = d[:, 1] / d[:, 1].max()
    elif eps_mode == "median":
        text += "Calculating eps using median\n"
        eps = np.sqrt(pow(np.median(d[:, 1]), 2) + pow(d[1, 0] - d[0, 0], 2))
        conf = d[:, 1] / d[:, 1].max()
    elif eps_mode == "range":
        text += "Calculating eps using range\n"
        eps = np.sqrt(
            pow((d[:, 1].max() - d[:, 1].min())*(1-args.tol), 2) + pow(d[1, 0] - d[0, 0], 2)
        )
        conf = d[:, 1] / d[:, 1].max()
    else:  # find distance using knee mehtod
        text += "Calculating eps using knee method\n"
        observation = int(len(amp) / 5)
        nbrs = NearestNeighbors(n_neighbors=observation).fit(d)
        # Find the k-neighbors of a point
        neigh_dist, _ = nbrs.kneighbors(d)
        # sort the neighbor distances (lengths to points) in ascending order
        sort_neigh_dist = np.sort(neigh_dist, axis=0)
        k_dist = sort_neigh_dist[:, observation - 1]  # sort_neigh_dist[:, 4]
        # plt.plot(k_dist)
        # plt.ylabel("k-NN distance")
        # plt.xlabel("Sorted observations (%ith NN)"%observation)
        # plt.show()
        kneedle = KneeLocator(
            x=range(1, len(neigh_dist) + 1),
            y=k_dist,
            S=1.0,
            curve="concave",
            direction="increasing",
            online=True,
        )
        eps = kneedle.knee_y
        conf = k_dist

    text += f"eps = [green]{eps:.4f}[/]    Minpoints = [green]{min_pts}[/]\n"
    model = DBSCAN(eps=eps, min_samples=min_pts)
    model.fit(d)
    dominant_index = model.labels_
    #normalize like the remaing methods
    dominant_index[dominant_index==-1] = dominant_index[dominant_index==-1] - 1 
    dominant_index =  dominant_index + 1 

    if "plotly" in args.engine:
        plot_outliers(args, freq_arr, amp, indecies, conf, dominant_index, d, eps)

    clean_index, _, msg = remove_harmonics(freq_arr, amp_tmp,  indecies[dominant_index == -1])
    dominant_index,text_d = dominant(clean_index, freq_arr, conf)
    
    return dominant_index, conf,  text+msg+text_d


# ?#################################
# ? Isolation Forest
# ?#################################
def isolation_forest(
    amp: np.ndarray, freq_arr: np.ndarray, args
) -> tuple[list[float], np.ndarray, str]:
    """calculates the outliers using isolation forest

    Args:
        amp (np.ndarray): amplitude or psd
        freq_arr (np.ndarray): freuqencies
        args (argsparse): arguments

    Returns:
        tuple[list[float], np.ndarray, str]: [dominant freuqency/ies, confidence]
    """
    text = "[green]Spectrum[/]: Amplitude spectrum\n" 
    if args.psd:
        amp = amp*amp/len(amp)
        text = "[green]Spectrum[/]: Power spectrum\n"

    indecies = np.arange(1, int(len(amp) / 2) + 1)
    amp_tmp = np.array(2 * amp[indecies])
    freq_arr_tmp = np.array(freq_arr[indecies])
    # norm the data
    # d = np.vstack((freq_arr_tmp / freq_arr_tmp.max(), amp_tmp / amp_tmp.max())).T
    #! norm over sum for amplitude with power spectrum 
    d = np.vstack((freq_arr_tmp / freq_arr_tmp.max(), amp_tmp / amp_tmp.sum())).T

    model = IsolationForest(contamination=0.001, warm_start=True)
    # model = IsolationForest(contamination=float(0.001),warm_start=True, n_estimators=2)
    # model = IsolationForest(warm_start=True)
    model.fit(d)
    conf = model.decision_function(d)
    dominant_index = model.predict(d)

    if "plotly" in args.engine:
        plot_outliers(args, freq_arr, amp, indecies, conf, dominant_index, d)
        plot_decision_boundaries(model,d, conf)

    clean_index, _, msg = remove_harmonics(freq_arr, amp_tmp,  indecies[dominant_index == -1])
    dominant_index,text_d = dominant(clean_index, freq_arr, conf)
    
    return dominant_index, abs(conf), text+msg+text_d


# ?#################################
# ? Odin
# ?#################################
def lof(amp: np.ndarray, freq_arr: np.ndarray, args) -> tuple[list[float], np.ndarray, str]:
    """calculates the outliers using isolation lof

    Args:
        amp (np.ndarray): amplitude or psd
        freq_arr (np.ndarray): freuqencies
        args (argsparse): arguments

    Returns:
        tuple[list[float], np.ndarray, str]: [dominant freuqency/ies, confidence]
    """
    text = "[green]Spectrum[/]: Amplitude spectrum\n" 
    if args.psd:
        amp = amp*amp/len(amp)
        text = "[green]Spectrum[/]: Power spectrum\n"

    indecies = np.arange(1, int(len(amp) / 2) + 1)
    amp_tmp = np.array(2 * amp[indecies])
    freq_arr_tmp = np.array(freq_arr[indecies])

    # norm the data
    # d = np.vstack((freq_arr_tmp / freq_arr_tmp.max(), amp_tmp / amp_tmp.max())).T
    #! norm over sum for amplitude with power spectrum 
    d = np.vstack((freq_arr_tmp / freq_arr_tmp.max(), amp_tmp / amp_tmp.sum())).T

    model = LocalOutlierFactor(contamination=0.001, novelty=True)
    model.fit(d)
    conf = model.decision_function(d)
    dominant_index = model.predict(d)
    # conf is between [-1,1]. Scale this to [0,1]
    conf = -1*(conf-1)/2

    # plot
    if "plotly" in args.engine:
        plot_outliers(args, freq_arr, amp, indecies, conf, dominant_index, d)

    clean_index, _, msg = remove_harmonics(freq_arr, amp_tmp,  indecies[dominant_index == -1])
    dominant_index,text_d = dominant(clean_index, freq_arr, conf)
    
    return dominant_index, abs(conf), text+msg+text_d


# ?#################################
# ? find_peaks
# ?#################################
def peaks(amp: np.ndarray, freq_arr: np.ndarray, args) -> tuple[list[float], np.ndarray, str]:
    """calculates the outliers using isolation lof

    Args:
        amp (np.ndarray): amplitude or psd
        freq_arr (np.ndarray): freuqencies
        args (argsparse): arguments

    Returns:
        tuple[list[float], np.ndarray, str]: [dominant freuqency/ies, confidence]
    """
    text = "[green]Spectrum[/]: Amplitude spectrum\n" 
    if args.psd:
        amp = amp*amp/len(amp)
        text = "[green]Spectrum[/]: Power spectrum\n"

    indecies = np.arange(1, int(len(amp) / 2) + 1)
    amp_tmp = np.array(2 * amp[indecies])
    freq_arr_tmp = np.array(freq_arr[indecies])
    
    # norm the data
    # d = np.vstack((freq_arr_tmp / freq_arr_tmp.max(), amp_tmp / amp_tmp.max())).T
    #! norm over sum for amplitude with power spectrum 
    d = np.vstack((freq_arr_tmp / freq_arr_tmp.max(), amp_tmp / amp_tmp.sum())).T

    limit = 1.2*np.mean(d[:,1])
    found_peaks, _ = find_peaks(d[:,1], height= limit if limit > 0.2 else 0.2)
    conf = np.zeros(len(d[:,1]))
    conf[found_peaks] = 1
    dominant_index = np.zeros(len(d[:,1]))
    dominant_index[found_peaks] = -1

    # plot
    if "plotly" in args.engine:
        plot_outliers(args, freq_arr, amp, indecies, conf, dominant_index, d)

    clean_index, _, msg = remove_harmonics(freq_arr, amp_tmp,  indecies[dominant_index == -1])
    dominant_index,text_d = dominant(clean_index, freq_arr, conf)
    
    return dominant_index, abs(conf), text+msg+text_d


def dominant(
    dominant_index: np.ndarray, freq_arr: np.ndarray, conf: np.ndarray
) -> tuple[list[float],str]:
    """_summary_

    Args:
        dominant_index (array): found indecies of dominant frequencies
        freq_arr (array): array of frequencies
        conf (float): value between -1 (strong outlier) and 1 (no outlier)

    Returns:
        dominant_index: set to 0 if more than three were found. Also remove harmonics
        text: text information
    """
    text = ""
    out = []
    if len(dominant_index) > 0:
        for i in dominant_index:
            if any(freq_arr[i] % freq_arr[out] < 0.00001):
                text += f"[yellow]Ignoring harmonic at: {freq_arr[i]:.3e} Hz (T = {1/freq_arr[i] if freq_arr[i] > 0 else 0:.3f} s, k = {i}) -> confidence: {abs(conf[i-1])*100:.3f}%[/]\n"
            else:
                text += f"Dominant frequency at: [green] {freq_arr[i]:.3e} Hz (T = {1/freq_arr[i] if freq_arr[i] > 0 else 0:.3f} s, k = {i}) -> confidence: {abs(conf[i-1])*100:.3f}%[/]\n"
                out.append(i)
            if len(out) > 2:
                text = "[red]Too many dominant frequencies -> Signal might be not perodic[/]\n"
                out = []
                break
    else: 
        text = "[red]No dominant frequencies found -> Signal might be not perodic[/]\n"
    
    return out, text


def remove_harmonics(freq_arr, amp_tmp, index_list) -> tuple[np.ndarray, list, str]:
    """Removes harmonics

    Args:
        freq_arr (_type_): frequency array
        amp_tmp (_type_): amplitude or power array
        index_list (_type_): list of indecies starting at 1

    Returns:
        np.ndarray: indecies without harmonics
        list: removed harmonics
        str: text to print
    """
    seen = []
    removed = []
    msg = ""
    flag = True
    #sort according to amplitude in descending order
    # index_list = index_list[np.argsort(-amp_tmp[index_list])]
    for ind in index_list:
        if seen:
            flag = True
            for value in seen:
                if freq_arr[ind] % freq_arr[value] < 0.00001 and ind != value:
                    msg += (
                        f"[yellow]Ignoring harmonic at: {freq_arr[ind]:.3e} Hz "
                        f"(T = {1/freq_arr[ind] if freq_arr[ind] > 0 else 0:.3f} s, k = {ind})[/]\n"
                        )
                    removed.append(ind)
                    flag = False
                    break
            if flag and ind not in seen:
                seen.append(ind)
        else:
            seen.append(ind)
    return np.array(seen), removed, msg



import numpy as np
import pandas as pd

def set_unit(arr:np.ndarray|float, suffix = "B/s") -> tuple[str, float]:
    """set unit for the plots

    Args:
        arr (np.ndarray): array
        unit (optional): default B/s

    Returns:
        unit (string): unit in GB/s, MB/s, KB/s or B/s
        arr: scaled array according to the unit
    """
    unit = suffix
    order = 1e-0
    if isinstance(arr, np.ndarray):
        pass
    elif isinstance(arr, pd.DataFrame):
        arr = arr.to_numpy()
    elif isinstance(arr,float) or isinstance(arr,int):
        arr = np.array(arr)
    
    if arr.size > 0 and np.max(arr) > 0:
        order = np.log10(np.max(arr))

    if order > 9:
        order = 1e-9
        unit = "G"+suffix
    elif order > 6:
        order = 1e-6
        unit = "M"+suffix
    elif order > 3:
        order = 1e-3
        unit = "K"+suffix
    else:
        pass

    return unit, order



def find_unit(df_t,index,index2,index_ind,index2_ind,args, offset=1):
    unit = "B/s"
    order = 1e-0
    if args.avr:
        tmp_unit, tmp_order = set_unit(df_t[1]["b_overlap_avr"][index][index2]*offset)
        if tmp_order < order:
            order = tmp_order
            unit = tmp_unit

    if args.sum:
        tmp_unit, tmp_order = set_unit(df_t[1]["b_overlap_sum"][index][index2]*offset)
        if tmp_order < order:
            order = tmp_order
            unit = tmp_unit

    if args.ind:
        tmp_unit, tmp_order = set_unit(df_t[3]["b_overlap_ind"][index_ind][index2_ind]*offset)
        if tmp_order < order:
            order = tmp_order
            unit = tmp_unit

    return unit, order


import numpy as np
from ftio.cli.ftio_core import core
from ftio.parse.args import parse_args
from ftio.freq._dft import display_prediction
from ftio.freq.freq_plot_core import convert_and_plot
from ftio.parse.bandwidth import overlap

ranks = 10
total_bytes = 100

# Set up data
## 1) overlap for rank level metrics
b_rank   = [0.0,0.0,1000.0,1000.0,0.0,0.0,1000.0,1000.0,0.0,0.0,1000.0,1000.0,0.0,0.0]
t_rank_s = [0.5,0.0,10.5,10.0,20.5,20.0,30.5,30.0,40.5,40.0,50.5,50.0,60.5,60]
t_rank_e = [5.0,4.5,15.0,14.5,25.0,24.5,35.0,34.5,45.0,44.5,55.0,54.5,65.0,64.5]
b,t = overlap(b_rank,t_rank_s, t_rank_e)

# ## 2) or directly specify the app level metrics
# t = [10.0, 20.1, 30.0, 40.2, 50.3, 60, 70, 80.0,]
# b = [10, 0, 10, 0, 10, 0, 10, 0]


# command line arguments
argv = ["-e", "no"] #["-e", "mat"]

# set up data
data = {
        "time": np.array(t),
        "bandwidth": np.array(b),
        "total_bytes": total_bytes,
        "ranks": ranks 
        }

#parse args
args = parse_args(argv,"ftio")

# perform prediction
prediction, dfs = core([data], args)


# plot and print info
convert_and_plot(data, dfs, args)
display_prediction("ftio", prediction)


"""Module contating function functions to discretize the data.
"""

from __future__ import annotations
import numpy as np


def sample_data(b: np.ndarray, t: np.ndarray, freq=-1) -> tuple[np.ndarray, float, str]:
    """Discretize the data according to the frequency

    Args:
        b (array): bandwidth
        t (array): time
        freq (int, optional): sampling frequency. Defaults to -1, calculates the optimal sampling
        frequency inside this function

    Returns:
        _type_: b_sampled, sampling frequency (in case -1), and text 
        
    """
    text = ""
    text += f"Time window: {t[-1]-t[0]:.2f} s\n"
    text += f"Frequency step: {1/(t[-1]-t[0]) if (t[-1]-t[0]) != 0 else 0:.3e} Hz\n"

    # ? calculate recommended frequency:
    if len(t) == 0:
        return np.empty(0), 0, " "
    if freq == -1:
        t_rec = np.inf
        for i in range(0, len(t) - 1):
            if (
                t_rec > (t[i + 1] - t[i])
                and (t[i + 1] - t[i]) != 0
                and (t[i + 1] - t[i]) >= 0.001
            ):
                t_rec = t[i + 1] - t[i]
        freq = 2 / t_rec
        text += f"Recommended sampling frequency: {freq:.3e} Hz\n"
    elif freq == -2:
        N = 10000
        freq = N/np.floor((t[-1] - t[0]))
    else:
        text += f"Sampling frequency:  {freq:.3e} Hz\n"
    N = int(np.floor((t[-1] - t[0]) * freq))
    text += f"Expected samples: {N}\n"
    # print("    '-> \033[1;Start time: %f s \033[1;0m"%t[0])

    # ? sample the data with the recommended frequency
    # t_sampled = np.zeros(N) #t[0]+np.arange(N)*1/freq
    b_sampled = np.zeros(N)
    n = len(t)
    counter = 0
    n_old = 0
    t_step = t[0]
    # error   = 0
    # errorStep   = 0
    for _ in range(0, N):
        for i in range(n_old, n):
            if (t_step >= t[i]) and (t_step < t[i + 1]) or i == n - 1:
                n_old = i  # no need to itterate over entire array
                b_sampled[counter] = b[i]
                counter = counter + 1
                break
        t_step = t_step + 1 / freq

    #! Abstraction error
    v_a = np.sum(np.abs(b_sampled * np.repeat(1 / freq, len(b_sampled))))
    v_0 = np.sum(b * (np.concatenate([t[1:], t[-1:]]) - t))
    # E = (abs(error))/(V0) if V0 > 0 else 0
    error = (abs(v_a - v_0)) / v_0 if v_0 > 0 else 0
    text += f"Abstraction error: {error:.5f}\n"

    if len(b_sampled) == 0:
        raise RuntimeError("No data in sampled bandwidth.\n Try increasing the sampling frequency")

    return b_sampled, freq, text[:-1]


def sample_data_same_size(b: np.ndarray, t:np.ndarray, freq=-1, n_bins=-1) -> tuple[np.ndarray,np.ndarray]:
    """Dsicretize the data according to the frequency and holds value

    Args:
        b (array): bandwidth
        t (array): time
        freq (int, optional): sampling frequency. Defaults to -1, calculates the optimal sampling
        frequency inside this function
        n_bins (int,optinal): number of desired bins

    Returns:
        _type_: b_sampled and sampling frequency (in case -1)
    """
    print(f"    '-> \033[1;34mBins: {n_bins}  \033[1;0m")
    b_sampled = np.zeros(n_bins)
    n_old = 0
    counter = 0
    t_step = 0
    n = len(b)
    for k in range(0, n_bins):
        if (t_step < t[0]) or (t_step > t[-1]):
            counter = counter + 1
        else:
            for i in range(n_old, n):
                if (t_step >= t[i]) and (t_step < t[i + 1]) or i == n - 1:
                    n_old = i  # no need to itterate over entire array
                    b_sampled[counter] = b[i]
                    counter = counter + 1
                    break
        t_step = t_step + 1 / freq

    t = np.arange(0, n_bins) * 1 / freq
    return b_sampled, t



""" Helper functions"""
from __future__ import annotations
import numpy as np
from rich.console import Console

def get_dominant(prediction: dict) -> float:
    """Gets the dominant frequency based on the confidence

    Args:
        prediction (dict): prediction contacting the dominant frequencies and their confidence

    Returns:
        float: dominant frequency (only one value!)
    """
    dominant_freq = prediction["dominant_freq"]
    conf = prediction["conf"]
    dominant_index = -1
    out = np.NaN
    if len(dominant_freq) != 0:
        dominant_index = np.argmax(conf)
        out = dominant_freq[dominant_index]

    return out


def get_dominant_and_conf(prediction: dict) -> tuple[float, float]:
    """Gets the dominant frequency and its confidence based on the confidence

    Args:
        prediction (dict): prediction contacting the dominant frequencies and their confidence

    Returns:
        tuple[float, float]: dominant frequency (only one value!) and corresponding confidence 
    """
    
    dominant_freq = prediction["dominant_freq"]
    conf = prediction["conf"]
    dominant_index = -1
    out_freq = np.NaN
    out_conf = np.NaN
    if len(dominant_freq) != 0:
        dominant_index = np.argmax(conf)
        out_freq = dominant_freq[dominant_index]
        out_conf = conf[dominant_index]

    return out_freq, out_conf


def print_data(data: list[dict]) -> None:
    """Prints the predictions in a nice format

    Args:
        data [dict]: contacting the predictions
    """
    print("Data collected is:")
    for pred in data:
        string = ""
        for keys, values in pred.items():
            if isinstance(values, float):
                string += f"'{keys}': {values:.2f}, "
            else:
                string += f"'{keys}': {values}, "
        print("{" + string[:-2] + "}")


def get_hits(prediction: dict, count: int, hits):
    """Manges the hit variable. In case a dominant frequency is found, hits is increased. 

    Args:
        prediction (dict): prediction up till now
        count (int): number of the prediction
        hits (Value): how often a dominant frequency was found

    Returns:
        hits: increased value if a dominant frequency was found, otherwise it is reset to 0
    """
    console = Console()
    text = ''
    text += f'[purple][PREDICTOR] (#{count}):[/] Freq candidates: \n'
    for i in range(0,len(prediction['dominant_freq'])):
        text += (
            f'[purple][PREDICTOR] (#{count}):[/]    {i}) '
            f'{prediction["dominant_freq"][i]:.2f} Hz -- conf {prediction["conf"][i]:.2f}\n'
        )
    if  len(prediction["dominant_freq"]) == 1:
        hits.value += 1
        text += f'[purple][PREDICTOR] (#{count}):[/] Current hits {hits.value}\n'
    else:
        hits.value = 0
        text += f'[purple][PREDICTOR] (#{count}):[/][red bold] Resetting hits {hits.value}[/]\n'

    console.print(text[:-1])

    return hits


def export_extrap(data: list[dict], name:str="./freq.jsonl"):
    """Generates measurement points for Extra-p out of the frequency 
    collected at different phases

    Args:
        data (list[dict]): List of predictions
        name (str, optional): Name of the output file. Defaults to "./freq.jsonl".
    """
    extrap_string, ranks = format_jsonl(data)

    if not np.isnan(ranks):
        name = f"./freq_{ranks}.jsonl"
    file = open(name, "w",encoding="utf-8")
    file.write(extrap_string)


def format_jsonl(data: list[dict]) -> tuple[str,str]:
    """Formats the metric as in the JSONL format for Extra-P

    Args:
        data (list[dict]): List of predictions

    Returns:
        tuple[str, str]: formatted string and number of ranks
        
    """
    string = ""
    out_ranks = np.NaN
    for pred in data:
        ranks = np.NaN
        call_path = "main"
        dominant_freq = np.NaN
        for keys, values in pred.items():
            if "dominant_freq" in keys:
                dominant_freq,_ = get_dominant_and_conf(pred)
            if "ranks" in keys:
                ranks = values
            if not np.isnan(dominant_freq) and not np.isnan(ranks):
                string +=f'{{"params":{{"Processes":{ranks}}},"callpath":"{call_path}","metric":"Frequency (Hz)","value":{dominant_freq:e} }}\n'
                out_ranks = ranks
                if dominant_freq > 0:
                    string +=f'{{"params":{{"Processes":{ranks}}},"callpath":"{call_path}","metric":"Period (s)","value":{1/dominant_freq:e} }}\n'
                break

    return string,out_ranks



def init_data(args:list) -> tuple[str, dict[str, dict|int], dict[str, int]]:
    """init data set

    Returns:
        tuple[dict[str, int], dict[str, int]]: empty data
    """
    
    io_data = {
        "number_of_ranks": 0,
        "total_bytes": 0,
        "max_bytes_per_rank": 0,
        "max_bytes_per_phase": 0,
        "max_io_phases_per_rank" : 0,
        "total_io_phases" : 0,
        "bandwidth": {
            "b_rank_sum": [],
            "b_rank_avr": [],
            "t_rank_s": [],
            "t_rank_e": [],
        },
    }
                    
    io_time = {
        "delta_t_sw": 0,
        "delta_t_sr": 0,
        "delta_t_awa": 0,
        "delta_t_awr": 0,
        "delta_t_aw_lost": 0,
        "delta_t_ara": 0,
        "delta_t_arr": 0,
        "delta_t_ar_lost": 0,
        "delta_t_overhead": 0,
        "delta_t_agg_io": 0,
        "delta_t_agg": 0
    }
    
    if isinstance(args, list):
        mode = "write_sync"
    else:
        if "w" in args.mode:
            mode =  "write"
        else:
            mode =  "read"
            
        if "async" in args.mode:
            mode = mode + "_async_t"
        else:
            mode = mode + "_sync"


    return mode, io_data, io_time





class Percent: 
    def  __init__(self,io_time): 
        # XXXX
        # |||'-> Bandwidth / Throughput / Delta
        # ||'--> Write / Read
        # |'---> Async / Sync / Overhead /  IO / Compute      
        # |----> Total / IO / Compute
        
        time = io_time.delta_t_agg     
        io = io_time.delta_t_agg_io  
        compute = io_time.delta_t_com     
        # [ ] fix this
# TODO: add flag to control granualaierty of sampling. Individual I/O operation can be discarded if focus is on phase (remove vectors)
# FIX: this is a todo

        # Async Write
        self.TAWB = self.percent(time, io_time.delta_t_awr)
        self.TAWT = self.percent(time, io_time.delta_t_awa)
        self.TAWD = self.percent(time, io_time.delta_t_aw_lost)
        self.IAWB = self.percent(io, io_time.delta_t_awr)
        self.IAWT = self.percent(io, io_time.delta_t_awa)
        self.IAWD = self.percent(io, io_time.delta_t_aw_lost)
        self.CAWB = self.percent(compute, io_time.delta_t_awr)
        self.CAWT = self.percent(compute, io_time.delta_t_awa)
        self.CAWD = self.percent(compute, io_time.delta_t_aw_lost)
    
        # Async Read
        self.TARB = self.percent(time, io_time.delta_t_arr)
        self.TART = self.percent(time, io_time.delta_t_ara)
        self.TARD = self.percent(time, io_time.delta_t_ar_lost)
        self.IARB = self.percent(io, io_time.delta_t_arr)
        self.IART = self.percent(io, io_time.delta_t_ara)
        self.IARD = self.percent(io, io_time.delta_t_ar_lost)
        self.CARB = self.percent(compute, io_time.delta_t_arr)
        self.CART = self.percent(compute, io_time.delta_t_ara)
        self.CARD = self.percent(compute, io_time.delta_t_ar_lost)

        # Sync Write
        self.TSW = self.percent(time, io_time.delta_t_sw)
        self.ISW = self.percent(io, io_time.delta_t_sw)
        self.CSW = self.percent(compute, io_time.delta_t_sw)
        
        # Sync Read
        self.TSR = self.percent(time, io_time.delta_t_sr)
        self.ISR = self.percent(io, io_time.delta_t_sr)
        self.CSR = self.percent(compute, io_time.delta_t_sr)
        
        # Lib overhead
        self.TO = self.percent(time, io_time.delta_t_overhead)
        self.IO = self.percent(io, io_time.delta_t_overhead)
        self.CO = self.percent(compute, io_time.delta_t_overhead)

        self.TI = self.percent(time, io)
        self.TC = self.percent(time, compute)
        self.CI = self.percent(compute, io)


    def percent(self,a,b):
        return b/a*100 if a != 0  else 0



from ftio.cli.ftio_core import main
from ftio.cli.ftio_core import core
from ftio.parse.args import parse_args
from ftio.freq._dft import display_prediction
from ftio.post.processing import label_phases

def test_ftio():
    args = parse_args(["-e", "no"], "ftio")
    _ = core([], args)
    assert True

def test_ftio_autocorrelation():
    args = parse_args(["-e", "no", "-c"], "ftio")
    _ = core([], args)
    assert True

def test_ftio_n_freq():
    file = "../examples/tmio/JSONL/8.jsonl"
    args = ["ftio", file, "-e", "no", "-n", "5"]
    _, args = main(args)
    assert True

def test_ftio_zscore():
    file = "../examples/tmio/JSONL/8.jsonl"
    args = ["ftio", file, "-e", "no"]
    prediction, args = main(args)
    assert prediction["t_start"] == 0.05309


def test_ftio_dbscan():
    file = "../examples/tmio/JSONL/8.jsonl"
    args = ["ftio", file, "-e", "no", "-o", "dbscan"]
    prediction, args = main(args)
    assert prediction["t_start"] == 0.05309


def test_ftio_lof():
    file = "../examples/tmio/JSONL/8.jsonl"
    args = ["ftio", file, "-e", "no", "-o", "lof"]
    prediction, args = main(args)
    assert prediction["t_start"] == 0.05309


def test_ftio_plot():
    file = "../examples/tmio/JSONL/8.jsonl"
    args = ["ftio", file, "-e", "no"]
    prediction, args = main(args)
    _ = label_phases(prediction, args)
    assert True 


def test_ftio_display_prediction():
    file = "../examples/tmio/JSONL/8.jsonl"
    args = ["ftio", file, "-e", "no"]
    prediction, args = main(args)
    display_prediction("ftio", prediction)
    assert True


import glob
from multiprocessing import Manager
from rich.console import Console
import ftio.prediction.monitor as pm
from ftio.prediction.helper import print_data, export_extrap
from ftio.prediction.async_process import handle_in_process
from ftio.prediction.probability_analysis import find_probability
from ftio.prediction.helper import get_dominant, get_hits
from ftio.api.gekkoFs.ftio_gekko import run
from ftio.prediction.analysis import display_result, save_data, data_analysis


def main(args: list[str] = []) -> None:

    n_buffers = 4
    args = ["-e", "plotly", "-f", "0.01"]
    # path=r'/d/github/FTIO/examples/API/gekkoFs/JSON/*.json'
    path = r"/d/github/FTIO/examples/API/gekkoFs/MSGPACK/write*.msgpack"
    # path = r"/tmp/gkfs_client_metrics/write*.msgpack"
    matched_files = glob.glob(path)

    # Init
    manager = Manager()
    queue = manager.Queue()
    data = manager.list()  # stores prediction
    aggregated_bytes = manager.Value("d", 0.0)
    hits = manager.Value("d", 0.0)
    start_time = manager.Value("d", 0.0)
    count = manager.Value("i", 0)
    procs = []

    # Init: Monitor a file
    stamp, _ = pm.monitor_list(matched_files, n_buffers)

    # Loop and predict if changes occur
    try:
        while True:
            # monitor
            stamp, procs = pm.monitor_list(matched_files, n_buffers, stamp, procs)

            # launch prediction_process
            procs.append(
                handle_in_process(
                    prediction_process,
                    args=(
                        data,
                        queue,
                        count,
                        hits,
                        start_time,
                        aggregated_bytes,
                        args,
                        matched_files,
                        n_buffers,
                    ),
                )
            )
    except KeyboardInterrupt:
        print_data(data)
        export_extrap(data=data)
        print("-- done -- ")


def prediction_process(
    data,
    queue,
    count,
    hits:int,
    start_time,
    aggregated_bytes:int,
    args: list[str],
    matched_files:list[str],
    n_buffers:int,
) -> None:
    console = Console()
    console.print(f"[purple][PREDICTOR] (#{count.value}):[/]  Started")
    # Modify the arguments
    args.extend(["-e", "no"])
    args.extend(["-ts", f"{start_time.value:.2f}"])

    # set up data
    if len(matched_files) != n_buffers:
        raise RuntimeError("Error, number of buffers does not match number of files")

    # Perform prediction
    prediction, args = run(matched_files, args)

    # get data
    freq = get_dominant(prediction)  # just get a single dominant value
    hits = get_hits(prediction, count.value, hits)

    # save prediction results
    save_data(queue, prediction, aggregated_bytes, count, hits)
    # display results
    text = display_result(freq, prediction, count, aggregated_bytes)
    # data analysis to decrease window
    text, start_time.value = data_analysis(args, prediction, freq, count, hits, text)
    console.print(text)
    count.value += 1
    while not queue.empty():
        data.append(queue.get())

    _ = find_probability(data)


if __name__ == "__main__":
    main()



from __future__ import annotations
import time
import numpy as np
from scipy.signal import find_peaks
from rich.panel import Panel
import plotly.graph_objects as go
# from rich.padding import Padding
import ftio.freq.discretize as dis
from ftio.freq.helper import format_plot
from ftio.freq.helper import MyConsole


# PLOT_MODE = "on"
# try:
# except ImportError:
#     PLOT_MODE = "empty"

CONSOLE = MyConsole()

def find_autocorrelation(args, data: dict, share:dict) -> dict:
    """Finds the period using autocorreleation

    Args:
        args (argparse): command line arguments
        data (dict): sampled data

    Returns:
        dict: predictions containing 4 fields: 
            1. bandwidth (np.array): bandwidth array
            2. time (np.array): time array indicating when bandwidth time points changed
            3. total_bytes (int): total transferred bytes
            4. ranks: number of ranks that did I/O
    """
    prediction = {}
    candidates = np.array([])
    fig = []
    CONSOLE.set(args.verbose)
    tik = time.time()
    if args.autocorrelation:
        CONSOLE.print("[cyan]Executing:[/] Autocorrelation\n")
        prediction = {
        "source":"autocorrelation",
        "dominant_freq": [],
        "conf": [],
        "t_start": 0,
        "t_end": 0,
        "total_bytes": 0,
        }
        # print("    '-> \033[1;32mExecuting Autocorrelation")
        b_sampled = np.array([])
        freq = np.NaN
        t_s = np.NaN
        t_e = np.NaN
        total_bytes = np.NaN
        # Ckeck if figure is on
        if any(x in args.engine for x in ["mat","plot"]):
            fig.append(go.Figure())

        # Take data if already avilable from previous step
        if share:
            b_sampled = share["b_sampled"]
            freq = share["freq"]
            t_s = share["t_start"]
            t_e = share["t_end"]
            total_bytes = share["total_bytes"]
        else:
            total_bytes = 0
            bandwidth = data["bandwidth"] if "bandwidth" in data else np.array([])
            time_b = data["time"] if "time" in data else np.array([])
            t_s = time_b[0]
            t_e = time_b[-1]
            if args.ts:  # shorten data
                time_b = time_b[time_b >= args.ts]
                t_s = time_b[0]
                bandwidth = bandwidth[len(bandwidth) - len(time_b) :]
                total_bytes = np.sum(
                    bandwidth * (np.concatenate([time_b[1:], time_b[-1:]]) - time_b)
                )
                CONSOLE.print(f"[purple]Start time set to {args.ts:.2f}")
            else:
                CONSOLE.print(f"[purple]Start time: {time_b[0]:.2f}")

            if args.te:  # shorten data
                time_b = time_b[time_b <= args.te]
                t_s = time_b[-1]
                bandwidth = bandwidth[len(bandwidth) - len(time_b) :]
                total_bytes = np.sum(
                    bandwidth * (np.concatenate([time_b[1:], time_b[-1:]]) - time_b)
                )
                CONSOLE.print(f"[purple]End time set to {args.te:.2f}")
            else:
                CONSOLE.print(f"[purple]End time: {time_b[-1]:.2f}")

            # sample the bandwidth bandwidth
            b_sampled, freq, text = dis.sample_data(bandwidth, time_b, args.freq) 
            CONSOLE.print(Panel.fit(text, style="white", border_style='yellow', title="Discretization", title_align='left'))

        # Scipy autocorrelation
        # lags = range(int(freq*len(b_sampled)))
        # acorr = sm.tsa.acf(b_sampled, nlags = len(lags)-1)
        #! numpy autocorrelation
        # # Mean
        mean = np.mean(b_sampled)
        # Variance
        var = np.var(b_sampled)
        # Normalized data
        ndata = b_sampled - mean
        # Caclualte autocorrelation:
        acorr = np.correlate(ndata, ndata, "full")[len(ndata) - 1 :]
        acorr = acorr / var / len(ndata)
        # plot
        if any(x in args.engine for x in ["mat","plot"]):
            fig[-1].add_scatter(y=acorr, mode="markers+lines", name="ACF", 
                                marker=dict(
                                            color = acorr,
                                            colorscale = ["rgb(0,50,150)", "rgb(150,50,150)", "rgb(255,50,0)"],
                                            showscale = True
                                            ))
            fig[-1].update_layout(
                font={"family": "Courier New, monospace", "size": 24, "color": "black"},
                xaxis_title="Lag (Samples)",
                yaxis_title="ACF",
                width = 1100,
                height = 400 #500
                )
            fig[-1].update_layout(coloraxis_colorbar=dict(yanchor="top", y=1, x=0, ticks="outside", ticksuffix=" bills"))
            # fig[-1].update_layout(legend=dict(yanchor="top", y=0.99, xanchor="right", x=0.01))
            fig[-1].update_layout(legend=dict(yanchor="top", y=0.99, xanchor="right", x=0.99))
            fig[-1] = format_plot(fig[-1])
            fig[-1].show()

        #! finding peak locations and calc the average time differences between them:
        peaks, prop  = find_peaks(acorr, height=0.15)
        candidates = np.diff(peaks) / freq
        weights = np.diff(prop['peak_heights'])
        # weights = []
        # heights = prop['peak_heights']
        # for i,_ in enumerate(heights):
        #     if i != len(heights) - 1:
        #         weights.append(abs(prop['peak_heights'][i+1]) +abs(prop['peak_heights'][i]))
        # weights = np.array(weights)
        # if len(weights) > 0:
        #     weights = weights/np.sum(weights)
        # print(peaks)
        # print(heights)
        # print(weights)
        
        if "mat" in args.engine or "plotly" in args.engine:
            fig[-1].add_scatter(
                x=peaks,
                y=acorr[peaks],
                marker=dict(
                    color = "rgba(20, 220, 70, 0.9)",
                    size=14, #12,
                    symbol="star-triangle-up",
                    angle=0,
                    line=dict(width=1, color="DarkSlateGrey")
                ),
                mode="markers",
                name="peaks",
            )

        # find outliers
        text  = ""
        outliers, text = filter_outliers(freq, candidates, weights)
        # remove outliers
        if len(outliers) != 0:
            candidates = np.delete(candidates, outliers)
            weights = np.delete(weights, outliers)
        # calculate period
        mean = np.average(candidates, weights=weights) if len(weights) > 0 else 0 
        std =  np.sqrt(np.abs(np.average((candidates-mean)**2, weights=weights))) if len(weights) > 0 else 0
        tmp = 1 / candidates if len(candidates) > 0 and any(candidates > 0) else 0
        if isinstance(tmp,list) and len(tmp) > 0:
            tmp = [f"{i:.4f}" for i in tmp]

        text += f"Found periods are [purple]{candidates}[/]\n"
        text += f"Matching frequencies are [purple]{tmp}[/]\n"
        periodicity = mean if len(candidates) > 0 else np.nan
        text += f"Average periods is [purple]{periodicity:.2f} [/]sec\n"
        text += f"Average frequency is [purple]{1/periodicity if periodicity > 0 else np.nan:.4f} [/]Hz\n"

        # calculate confidence using "Coefficient of variation": https://en.wikipedia.org/wiki/Coefficient_of_variation
        coef_var = np.abs(std / mean ) if len(candidates) > 0 else np.nan
        # coef_var = np.abs(np.std(candidates) / np.mean(candidates) ) if len(candidates) > 0 else np.nan
        conf = np.abs(1 - coef_var)
        text += f"confidence is [purple]{conf*100:.2f} [/]%\n"
        
        # CONSOLE.print(Padding(Panel.fit(text[:-1], style="white", border_style='purple', title="Autocorreleation", title_align='left'), (0, 4)))
        CONSOLE.print(Panel.fit(text[:-1], style="white", border_style="purple", title="Autocorreleation", title_align="left"))

        if "mat" in args.engine or "plotly" in args.engine:
            if len(candidates) > 0:
                val = np.delete(peaks,outliers)
                fig[-1].add_scatter(
                    x=val,
                    y=acorr[val],
                    marker=dict(
                        color = "rgba(220, 20, 70, 0.9)",
                        size=21, #19,
                        symbol="circle-open-dot",
                        angle=0,
                        line=dict(width=2, color="DarkSlateGrey")
                    ),
                    mode="markers",
                    name="relevant peaks",
                )
            fig[-1] = format_plot(fig[-1])
            fig[-1].show(config = {"toImageButtonOptions": {"format": "png", "scale": 5}})

        prediction["dominant_freq"] = 1/periodicity  if periodicity > 0 else np.nan
        prediction["conf"] = conf
        prediction["t_start"] = t_s
        prediction["t_end"] = t_e
        prediction["total_bytes"] = total_bytes
        prediction["freq"] = freq
        prediction["candidates"] = candidates
        CONSOLE.print(f"\n[cyan]Autocorrelation finished:[/] {time.time() - tik:.3f} s")
        
    return prediction


def filter_outliers(freq: float, candidates: np.ndarray, weights: np.ndarray) -> tuple[np.ndarray, str]:
    """removes outliers using either qunatil method or Z-score

    Args:
        candidates (np.ndarray): peaks
        weights (np.ndarray): weights of the peaks

    Returns:
        np.ndarray: outliers
        str: string text
    """
    text = ""
    outliers = np.array([])
    # remove outliers:
    if len(candidates) > 0 and any(candidates > 0):
        ind = np.where(candidates > 1/freq)
        candidates = candidates[ind] # remove everythin above 10 Hz
        if len(weights > 0):
            weights = weights[ind]
        method = "z"
        # With quantil and weights
        if "q" in method:
            text += f"Filtering method: [purple]quantil[/]\n"
            # candidates = candidates*weights/sum(weights)
            q1 = np.percentile(candidates, 25)
            q3 = np.percentile(candidates, 75)
            iqr = q3 - q1
            threshold = 1.5 * iqr
            outliers = np.where(
                (candidates < q1 - threshold) | (candidates > q3 + threshold)
            )
        elif "z" in method:
            text += "Filtering method: [purple]Z-score with weighteed mean[/]\n"
            # With Zscore:
            mean = np.average(candidates, weights=weights) if len(weights) > 0 else 0
            # std = np.std(candidates)
            std =  np.sqrt(np.abs(np.average((candidates-mean)**2, weights=weights))) if len(weights) > 0 else 0
            text += f"Wighted mean is [purple]{mean:.3f}[/] and weighted std. is [purple]{std:.3f}[/]\n"
            z_score = np.abs((candidates - mean) / std) if std != 0  else np.array([])
            outliers = np.where(z_score > 1)
            text += f"Z-score is [purple]{print_array(z_score)}[/]\n"
        
        text += (
            f"[purple]{len(candidates)}[/] period candidates found:\n"
            f"[purple]{print_array(candidates)}[/]\n\n"
            f"{len(candidates[outliers])} outliers found:\n[purple]{print_array(candidates[outliers])}[/]\n\n"
        )
    else:
        text += "[purple]Empty[/]\n"

    return outliers, text

def print_array(array:np.ndarray) -> str:
    out = ""
    if len(array) == 0:
        out =" "
    for i in array:
        out += f" {i:.2f}"
    

    return "["+out[1:]+"]"



from __future__ import annotations
import re 
import statistics as st
from ftio.parse.sample import Sample
from ftio.parse.time_io import Time
from ftio.parse.percent import Percent
from ftio.parse.helper import match_mode
class Simrun:
    """Stores the result from a single simulation including the:
    1. read sync
    2. read async
    3. write sync
    4. write async
    5. I/O time (if available)
    6. I/O percent (if available)
    """
    def __init__(self, data, ext: str,  name: str, args, file_index: int = 0):
        """create Simrun object

        Args:
            data (array): contains data from simulation
            ext (str): 'json', 'jsonl', 'recorder', 'darshan', 'msgpack', or 'txt'
            name (str): name of simulation (e.g., 192.json)
            args (argparse): command line arguments
            file_index (int, optional): file index, in case several files have the same name
        """
        self.name       = name
        self.ranks      = self.get_rank(ext)
        mode            = match_mode(args.mode) if args.mode else ''
        args.file_index = file_index
        supported_modes = ['read_sync', 'write_sync', 'read_async', 'write_async']
        
        #! list = JSONL
        if isinstance(data,list):
            #! JSONL need to be converted to a dict
            # if 'Time' in data[-1]:
            if 'jsonl' in ext or 'msgpack' in ext:
                if mode and mode in supported_modes:
                    self.reset(args)
                    self.assign(data, args, mode,'jsonl')
                else:
                    self.read_sync     = self.merge_parts(data,'read_sync', args)
                    self.write_sync    = self.merge_parts(data,'write_sync', args)
                    self.read_async_t  = self.merge_parts(data,'read_async_t', args)
                    self.read_async_b  = self.merge_parts(data,'read_async_b', args)
                    self.write_async_t = self.merge_parts(data,'write_async_t', args)
                    self.write_async_b = self.merge_parts(data,'write_async_b', args)
                    if any('io_time' in d for d in data):
                        self.io_time       = self.merge_parts(data,'io_time',args)
                    else:
                        self.io_time       = Time({},self.ranks,args)
                    self.io_percent    = Percent(self.io_time)
            else:
                raise ValueError('Data format empty or not supported')

        #! Darshan files or recorder files
        elif 'darshan' in ext or 'recorder' in ext:
            self.reset(args)
            self.read_sync         = Sample(data['read_sync'],'read_sync',args)
            self.write_sync        = Sample(data['write_sync'],'write_sync',args)
            self.io_time           = Time(data['io_time'],self.ranks,args)

        elif 'txt' in ext:
            self.reset(args)
            setattr(self,mode,Sample(data[mode],mode,args))
            self.io_time           = Time(data['io_time'],self.ranks,args)

        #! json files
        else:
            #! for dft to make it faster
            if mode and mode in supported_modes:
                self.reset(args)
                self.assign(data,args,mode)
                return
            #! standard json files
            self.read_sync     = Sample(data['read_sync'],    'read_sync',args)
            self.write_sync    = Sample(data['write_sync'],   'write_sync',args)
            self.read_async_t  = Sample(data['read_async_t'], 'read_async_t',args)
            self.read_async_b  = Sample(data['read_async_b'], 'read_async_b',args)
            self.write_async_t = Sample(data['write_async_t'],'write_async_t',args)
            self.write_async_b = Sample(data['write_async_b'],'write_async_b',args)
            if 'io_time' in data:
                self.io_time       = Time(data['io_time'],self.ranks,args)
                self.io_percent    = Percent(self.io_time)


#---------------------------------------------------------------------------------------------------
#?=======================
#? helper functions
#?=======================
    def assign(self,data, args, mode, file_format = 'json'):
        """assigns a single mode instead of parsing all data

        Args:
            data (array): sim data
            args (argparse): command line arguments
            mode (str): mode to assign
            format (str, optional): json or jsonl. Defaults to 'json'.
        """
        if 'jsonl' in file_format:
            if 'read_sync' == mode:
                self.read_sync     = self.merge_parts(data,'read_sync',args)
            elif 'read_async' == mode:
                self.read_async_t  = self.merge_parts(data,'read_async_t',args)
                self.read_async_b  = self.merge_parts(data,'read_async_b',args)
            elif 'write_sync' == mode:
                self.write_sync     = self.merge_parts(data,'write_sync',args)
            elif 'write_async' == mode:
                self.write_async_t  = self.merge_parts(data,'write_async_t',args)
                self.write_async_b  = self.merge_parts(data,'write_async_b',args)
            else:
                pass

            if any('io_time' in d for d in data):
                self.io_time       = self.merge_parts(data,'io_time',args)
                self.io_percent    = Percent(self.io_time)

        else:
            if 'read_sync' == mode:
                self.read_sync     = Sample(data['read_sync'],    'read_sync',args)
            elif 'read_async' == mode:
                self.read_async_t  = Sample(data['read_async_t'], 'read_async_t',args)
                if 'read_async_b' in data:
                    self.read_async_b  = Sample(data['read_async_b'], 'read_async_b',args)
            elif 'write_async' == mode:
                self.write_async_t = Sample(data['write_async_t'],'write_async_t',args)
                if 'write_async_b' in data:
                    self.write_async_b = Sample(data['write_async_b'],'write_async_b',args)
            elif 'write_sync' == mode:
                self.write_sync    = Sample(data['write_sync'],   'write_sync',args)


    def reset(self, args):
        """sets all fields to empty. This is usually followed by a assign call

        Args:
            args (argparse): command line arguments
        """
        self.read_sync     = Sample({'bandwidth':[]},'read_sync', args)
        self.read_async_t  = Sample({'bandwidth':[]},'read_async_t', args)
        self.read_async_b  = Sample({'bandwidth':[]},'read_async_b', args)
        self.write_async_t = Sample({'bandwidth':[]},'write_async_t', args)
        self.write_async_b = Sample({'bandwidth':[]},'write_async_b', args)
        self.write_sync    = Sample({'bandwidth':[]},'write_sync', args)
        self.io_time       = Time({}, self.ranks, args)


    def get_rank(self,ext:str) -> int:
        """Get rank name from extension

        Args:
            ext (str): extension type

        Returns:
            int: number of ranks that did I/O
        """
        out = -1
        if isinstance(self.name,int):
            out = self.name
        elif isinstance(self.name,str) and '.' not in self.name and '/' not in self.name:
            out = int(self.name)
        else:
            try:
                name_s = self.name.rfind("/")
                if name_s < 0:
                    name_s = self.name.rfind("\\") # for windows
                name_e = self.name.rfind(f".{ext}")
                rank = self.name[name_s+1:name_e]
                strs = ["_","-"," "]
                if any(x in rank for x in strs):
                    for x in strs:
                        if x in rank:
                            name_e = rank.rfind(x)
                            rank = rank[:name_e]
                out = int(rank)
            except ValueError:
                rank = re.findall(r'\d+', "hello 42 I'm a 32 string 30")
                out = int(rank[-1])
        return out


    def print_rank(self,file):
        """print rank mapped to POINTS. This is used for 
        Extra-P with the text file format.

        Args:
            file (fileptr): file pointer
        """
        file.write(f"POINTS ( {self.ranks:i} )\n")


    def merge_parts(self, data, mode, args):
        """merge jsonl parts to a single simulation

        Args:
            data (2d array): array fo data elements
            mode (str): mode to merge
            args (argparse): command line arguments

        Returns:
            io_sample object: iosample of the whole simulation
        """
        data_array     = [item[mode] for item in data if mode in item]
        out = []
        if len(data_array) <= 1:
            if "time" in mode:
                out = Time(data_array[0],self.ranks,args) if len(data_array) > 0 else Time({},self.ranks,args)
            else:
                out = Sample(data_array[0],mode,args)  if len(data_array) > 0 else Sample({},mode,args)
        else:
            data_array = self.merge_fields(data_array)
            if "time" in mode:
                out = Time(data_array,self.ranks,args)
            else:
                out = Sample(data_array,mode,args)
        return out


    def merge_fields(self,data_array, keys: list[str] = []) -> dict:
        """Merges the metrics field from different files. For example,
        JsonlLines file constantly append new data. To merge the previous 
        metrics with the new one, this function iterates over the fields 
        and merges them.

        Args:
            data_array (dataframe): Metrics
            keys (list[str], optional): _description_. Defaults to list[str].

        Returns:
            dict: merge fields in stored in a dict variable
        """
        if not keys:
            keys = [k for k in data_array[0].keys()]

        my_dict = {}
        for field in keys:
            # print(f)
            if isinstance(data_array[0][field],dict):
                data_array2  = [x[field] for x in data_array if field in x]
                my_dict[field] = self.merge_fields(data_array2,[k for k in data_array2[0].keys()])
            else:
                if isinstance(data_array[0][field],list):
                    my_dict[field] = []
                    for i,_ in enumerate(data_array):
                        my_dict[field].extend(data_array[i][field])
                else:
                    if any([x in field for x in ['total','_t_']]):
                        my_dict[field] = sum(x[field] for x in data_array)
                    elif any([x in field for x in ['max','number']]):
                        my_dict[field] = max(x[field] for x in data_array)
                    elif 'min' in field:
                        my_dict[field] = min(x[field] for x in data_array)
                    elif 'arithmetic_mean' in field:
                        my_dict[field] = st.mean(x[field] for x in data_array)

        return my_dict


"""class to parse the data

    Returns:
    Sample: contains metrics for a single I/O type (read/write + async/sync)
"""
from ftio.parse.bandwidth import Bandwidth


class Sample:
    """contains metrics for a single I/O type (read/write + async/sync)
    """

    def __init__(self, values: dict, io_type:str, args) -> None:
        self.type                        = io_type
        self.max_bytes_per_rank          = self.assign(values,"max_bytes_per_rank")
        #self.max_offset_over_ranks       = self.assign(values,"max_offset_over_ranks")
        self.total_bytes                 = self.assign(values,"total_bytes")
        self.max_bytes_per_phase         = self.assign(values,"max_bytes_per_phase")
        self.max_io_phases_per_rank      = self.assign(values,"max_io_phases_per_rank")
        self.total_io_phases             = self.assign(values,"total_io_phases")
        self.max_io_ops_per_rank         = self.assign(values,"max_io_ops_per_rank")
        self.max_io_ops_in_phase         = self.assign(values,"max_io_ops_in_phase")
        self.total_io_ops                = self.assign(values,"total_io_ops")
        self.number_of_ranks             = self.assign(values,"number_of_ranks")
        self.bandwidth                   = self.assign_bandwidth(values, io_type, args)
        self.file_index                  = args.file_index

    def get_data(self):
        #! append list is much faster than append data frame
        name0 = []
        data0 = []
        common        = ['number_of_ranks','file_index']
        name_rank_ovr = ['b_overlap_sum','b_overlap_avr', 't_overlap']
        name_rank     = ['b_rank_sum', 'b_rank_avr', 't_rank_s', 't_rank_e']
        name_ind_ovr  = ['b_overlap_ind', 't_overlap_ind']
        name_ind      = [ 'b_ind', 't_ind_s','t_ind_e']

        # add phase statistics
        for attr, value in self.__dict__.items():
            if (attr not in ['bandwidth', 'file_index']):
                name0.append(attr)
                data0.append(value)

        # statistics:
        name_tmp = common + name_rank_ovr + name_rank + name_ind_ovr + name_ind
        for attr, value in self.bandwidth.__dict__.items():
            if (attr not in name_tmp):
                name0.append(attr)
                data0.append(value)

        name_rank_ovr, data1 = self.find_data(name_rank_ovr,common)
        name_rank,     data2 = self.find_data(name_rank,common)
        name_ind_ovr,  data3 = self.find_data(name_ind_ovr,common)
        name_ind,      data4 = self.find_data(name_ind,common)

        return name0, [data0], name_rank_ovr, data1, name_rank, data2, name_ind_ovr, data3, name_ind, data4


    def assign(self,values, name):
        if name in values:
            return values[name]
        else:
            return float('NaN')
    
    def assign_bandwidth(self, values, name, args):
        if "bandwidth" in values:
            return Bandwidth(values["bandwidth"], name, args)
        else:
            return Bandwidth({}, name, args)

    def find_data(self,name,common):
        #remove empty:
        for i in name:
            if not getattr(self.bandwidth,i):
                name.remove(i)

        # add phase info
        data = [getattr(self.bandwidth,s) for s in name]
        l = len(data[-1])
        name = name + common
        data.extend([[self.number_of_ranks]*l,[self.file_index]*l])
        return name, data


"""Legendgroup names and there titles for plotly plots"""

# Groups
# actual
ACTUAL_AVERAGE: str = "actual_average"
ACTUAL_SUM: str = "actual_sum"
ACTUAL_INDIVIDUAL: str = "actual_individual"
# required
REQUIRED_AVERAGE: str = "required_average"
REQUIRED_SUM: str = "required_sum"
REQUIRED_INDIVIDUAL: str = "required_individual"

# Titles
ACTUAL_AVERAGE_TITLE: str = "Actual average"
ACTUAL_SUM_TITLE: str = "Actual sum"
ACTUAL_INDIVIDUAL_TITLE: str = "Actual individual"
# required
REQUIRED_AVERAGE_TITLE: str = "Required average"
REQUIRED_SUM_TITLE: str = "Required sum"
REQUIRED_INDIVIDUAL_TITLE: str = "Required individual"



from dash import dcc, html
from dash_extensions.enrich import DashProxy, ServersideOutputTransform, TriggerTransform

import ftio.plot.dash_files.constants.id as id
import ftio.plot.dash_files.constants.io_mode as io_mode
from ftio.plot.dash_files.callback_files.callbacks import get_callbacks


class IOAnalysisApp(DashProxy):
    def __init__(self, plot_core) -> None:
        super().__init__(__name__, transforms=[ServersideOutputTransform(), TriggerTransform()])
        self.title = "IO Analysis"

        self._plot_core = plot_core
        self._io_modes = self._collect_io_modes()

        self.layout = self._div_layout()

        get_callbacks(self, self._plot_core, self._io_modes)

    def _div_layout(self) -> html.Div:
        return html.Div(
            children=[
                dcc.Loading(
                    [
                        dcc.Store(id=id.STORE_FIGURES_SYNC_READ, data={}, storage_type="memory"),
                        dcc.Store(id=id.STORE_FIGURES_SYNC_WRITE, data={}, storage_type="memory"),
                        dcc.Store(id=id.STORE_FIGURES_ASYNC_READ, data={}, storage_type="memory"),
                        dcc.Store(id=id.STORE_FIGURES_ASYNC_WRITE, data={}, storage_type="memory"),
                        # dcc.Store(id=id.STORE_FIGURES_TIME, data={}, storage_type="memory"),
                    ],
                    fullscreen=True,
                    type="dot",
                ),
                html.H1(self.title),
                html.Hr(),
                html.Div(children="Number of files: " + str(len(self._plot_core.data.paths))),
                html.Hr(),
                dcc.Dropdown(
                    options=self._plot_core.data.paths,
                    value=self._plot_core.data.paths,
                    id=id.DROPDOWN_FILE,
                    multi=True,
                    style={"marginTop": 10},
                    disabled=True if self._plot_core.data.args.merge_plots else False,
                ),
                dcc.Checklist(
                    options=self._io_modes,
                    value=self._io_modes,
                    inline=True,
                    style={"marginBottom": 10, "marginTop": 10},
                    id=id.CHECKLIST_IO_MODE,
                ),
                html.Button("Show", id=id.BUTTON_SHOW, n_clicks=0),
                html.Hr(),
                html.Div(id=id.DIV_SYNC_READ, children=[]),
                html.Div(id=id.DIV_SYNC_WRITE, children=[]),
                html.Div(id=id.DIV_ASNYC_READ, children=[]),
                html.Div(id=id.DIV_ASYNC_WRITE, children=[]),
                # html.Div(id=id.DIV_IO_TIME, children=[]),
            ],
        )

    def _collect_io_modes(self) -> list[str]:
        data = self._plot_core.data
        io_modes = []
        if not (data.df_rat is None or data.df_rat[1].empty) or not (
            data.df_rab is None or data.df_rab[1].empty
        ):
            io_modes.append(io_mode.ASYNC_READ)
        if not (data.df_wat is None or data.df_wat[1].empty) or not (
            data.df_wab is None or data.df_wab[1].empty
        ):
            io_modes.append(io_mode.ASYNC_WRITE)
        if not (data.df_rst is None or data.df_rst[1].empty):
            io_modes.append(io_mode.SYNC_READ)
        if not (data.df_wst is None or data.df_wst[1].empty):
            io_modes.append(io_mode.SYNC_WRITE)

        return io_modes



"""Performs the analysis for prediction. This includes the calculation of ftio and parsing of the data into a queue """
from __future__ import annotations
from multiprocessing import  Queue
from rich.console import Console
import numpy as np
from ftio.cli import ftio_core
from ftio.prediction.helper import get_dominant, get_hits
from ftio.plot.units import set_unit

def ftio_process(queue: Queue, count, hits, start_time, aggregated_bytes, args, msgs=None) -> None:
    """Perform a single prediction

    Args:

        queue (Manager().Queue): queue for FTIO data
        count (Manager().Value): number of prediction
        hits (Manager().Value): hits indicating how often a dominant frequency was found
        start_time (Manager().Value): start time window for ftio
        aggregated_bytes (Manager().Value): total bytes transferred so far
        args (list[str]): additional arguments passed to ftio
    """
    console = Console()
    console.print(f'[purple][PREDICTOR] (#{count.value}):[/]  Started')
    # Modify the arguments
    args.extend(['-e', 'no'])
    args.extend(['-ts', f'{start_time.value:.2f}'])
    
    # perform prediction
    prediction, args = ftio_core.main(args,msgs)
    
    # get data
    freq = get_dominant(prediction) #just get a single dominant value
    hits = get_hits(prediction,count.value,hits)

    # save prediction results
    save_data(queue, prediction, aggregated_bytes, count, hits)
    # display results
    text = display_result(freq ,prediction ,count, aggregated_bytes)
    # data analysis to decrease window
    text, start_time.value = data_analysis(args, prediction, freq, count, hits, text)
    console.print(text)
    count.value += 1


def data_analysis(args, prediction, freq, count, hits, text:str) -> tuple[str, float]:
    # average data/data processing
    t_s = prediction['t_start']
    t_e = prediction['t_end']
    total_bytes = prediction['total_bytes']
    if not np.isnan(freq):
        n_phases = (t_e-t_s)*freq
        avr_bytes = int(total_bytes/float(n_phases))
        unit, order = set_unit(avr_bytes,'B')
        avr_bytes = order *avr_bytes
        
        #FIXME this needs to compensate for a smaller windows
        if not args.window_adaptation:
            text += (
                f'[purple][PREDICTOR] (#{count.value}):[/] Estimated phases {n_phases:.2f}\n'
                f'[purple][PREDICTOR] (#{count.value}):[/] Average transferred {avr_bytes:.0f} {unit}\n'
            )
        
        # adaptive time window
        if args.window_adaptation:
            if hits.value > args.frequency_hits: 
                if True: #np.abs(avr_bytes - (total_bytes-aggregated_bytes.value)) < 100:
                    tmp = t_e - 3*1/freq
                    t_s = tmp if tmp > 0 else 0
                    text += f'[purple][PREDICTOR] (#{count.value}):[/][green] Adjusting start time to {t_s} sec\n[/]'
            else:
                t_s = 0
                if hits.value == 0:
                    text += f'[purple][PREDICTOR] (#{count.value}):[/][red bold] Resetting start time to {t_s} sec\n[/]'
                
    # TODO 1: Make sanity check -- see if the same number of bytes was transferred
    # TODO 2: Train a model to validate the predictions?
    text += f'[purple][PREDICTOR] (#{count.value}):[/] Ended'

    return text, t_s


def save_data(queue, prediction, aggregated_bytes, count, hits) -> None:
    """Put all data from `prediction` in a `queue`. The total bytes are as well saved here. 

    Args:
        queue (Manager.queue): queue containing the data from the prediction 
        prediction (dict): result from FTIO
        aggregated_bytes (_type_): total bytes transferred over entire runtime 
        count (_type_): prediction number
        hits (_type_): home many times a dominant frequency was found
    """
    # safe total transferred bytes
    aggregated_bytes.value += prediction['total_bytes']
    
    # save data
    queue.put(
        {
            'phase': count.value,
            'dominant_freq': prediction['dominant_freq'],
            'conf': prediction['conf'],
            'amp': prediction['amp'],
            'phi': prediction['phi'],
            't_start': prediction['t_start'],
            't_end': prediction['t_end'],
            'total_bytes': prediction['total_bytes'],
            'ranks': prediction['ranks'],
            'freq': prediction['freq'],
            'hits': hits.value,
        }
    )


def display_result(freq: float ,prediction: dict ,count, aggregated_bytes) -> str:
    text = ''
    if not np.isnan(freq):
        text = f'[purple][PREDICTOR] (#{count.value}):[/] Dominant freq {freq:.3f} \n'

    # time window
    text += (
        f'[purple][PREDICTOR] (#{count.value}):[/] Time window {prediction["t_end"]-prediction["t_start"]:.3f} sec ([{prediction["t_start"]:.3f},{prediction["t_end"]:.3f}] sec)\n')

    # total bytes
    total_bytes = aggregated_bytes.value
    # total_bytes =  prediction["total_bytes"]
    unit, order = set_unit(total_bytes,'B')
    total_bytes = order*total_bytes
    text += (
        f'[purple][PREDICTOR] (#{count.value}):[/] Total bytes {total_bytes:.0f} {unit}\n')

    # Bytes since last time
    # tmp = abs(prediction["total_bytes"] -aggregated_bytes.value)
    tmp = abs(aggregated_bytes.value)
    unit, order = set_unit(tmp,'B')
    tmp = order *tmp
    text += (
        f'[purple][PREDICTOR] (#{count.value}):[/] Bytes transferred since last '
        f'time {tmp:.0f} {unit}\n'
        )

    return text